# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc

from tikka.interfaces.adapters.network.indexer.indexer import NetworkIndexerInterface


class IndexerAccountsInterface(abc.ABC):
    """
    IndexerAccountsInterface class
    """

    def __init__(self, indexer: NetworkIndexerInterface) -> None:
        """
        Init IndexerAccountsInterface instance

        :param indexer: NetworkIndexerInterface instance
        :return:
        """
        self.indexer = indexer

    @abc.abstractmethod
    def is_legacy_v1(self, address: str) -> bool:
        """
        Return True if account is a legacy v1 account (present in genesis)

        :param address: ss58 account address
        :return:
        """
        raise NotImplementedError


class IndexerAccountsException(Exception):
    """
    IndexerAccountsException class
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable

from PyQt5.QtCore import QObject, pyqtSignal

from tikka.domains.entities.account import Account
from tikka.domains.entities.category import Category
from tikka.domains.entities.identity import Identity


class EventBus(QObject):
    """
    EventBus class
    """

    # singleton pattern
    _instance = None

    #########################################
    # SIGNALS ###############################
    #########################################

    # main window
    main_window_shown = pyqtSignal()
    repository_data_updated = pyqtSignal()
    user_data_restored = pyqtSignal()

    # categories
    category_added = pyqtSignal(Category)
    category_deleted = pyqtSignal(Category)

    # accounts
    account_added = pyqtSignal(Account)
    account_forgotten = pyqtSignal(Account)
    account_renamed = pyqtSignal(Account)
    account_wallet_updated = pyqtSignal(Account)
    account_selected = pyqtSignal(Account)
    account_balances_updated = pyqtSignal(list)
    account_identities_updated = pyqtSignal(list)
    account_profiles_updated = pyqtSignal(list)

    # transfers
    transfer_sent = pyqtSignal(list)

    # identities
    identity_created = pyqtSignal(Identity, Identity)
    identity_confirmed = pyqtSignal(Identity)
    identity_certified = pyqtSignal(Identity, Identity)
    identity_account_changed = pyqtSignal(Account, Account)
    identities_updated = pyqtSignal(list)

    # profile cesium+
    profile_account_changed = pyqtSignal(Account, Account)

    # currency
    currency_changed = pyqtSignal(str, str)
    currency_closed = pyqtSignal(str)
    currency_opened = pyqtSignal(str)

    # servers
    node_connected = pyqtSignal()
    indexer_connected = pyqtSignal()
    datapod_connected = pyqtSignal()
    node_disconnected = pyqtSignal()
    indexer_disconnected = pyqtSignal()
    datapod_disconnected = pyqtSignal()

    @classmethod
    def instance(cls, parent=None):
        """
        Initialize Singleton instance

        :param parent: QObject instance
        :return:
        """
        if cls._instance is None:
            cls._instance = cls(parent)
        return cls._instance

    def __init__(self, parent=None):
        """
        Initialize EventBus instance

        :param parent: QObject instance
        """
        super().__init__(parent)

    def connect(self, signal: pyqtSignal, slot: Callable):
        """
        Debug connect

        :param signal: pyqtSignal instance
        :param slot: Callable
        :return:
        """
        print(f"DEBUG: Connection {signal} -> {slot}")
        result = signal.connect(slot)  # type: ignore
        return result

    def emit(self, signal, *args):
        """
        Debug emit

        :param signal: pyqtSignal instance
        :param args: Arguments
        :return:
        """
        print(f"DEBUG: Emit signal {signal} with args: {args}")
        print(f"DEBUG: Receivers: {signal.receivers()}")
        signal.emit(*args)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
from typing import Any, List, Optional

from tikka.domains.entities.profile import Profile
from tikka.interfaces.adapters.repository.db_repository import DBRepositoryInterface
from tikka.interfaces.adapters.repository.profiles import ProfilesRepositoryInterface

TABLE_NAME = "profiles"


class DBProfilesRepository(ProfilesRepositoryInterface, DBRepositoryInterface):
    """
    DBProfilesRepository class
    """

    def add(self, profile: Profile) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ProfilesRepositoryInterface.add.__doc__
        )

        # insert only non hidden fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_profile(profile),
        )

    def get(self, address: str) -> Optional[Profile]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ProfilesRepositoryInterface.get.__doc__
        )

        row = self.client.select_one(
            f"SELECT * FROM {TABLE_NAME} WHERE address=?", (address,)
        )
        if row is None:
            return None

        return get_profile_from_row(row)

    def update(self, profile: Profile) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ProfilesRepositoryInterface.update.__doc__
        )

        # update only non-hidden fields
        self.client.update(
            TABLE_NAME,
            f"address='{profile.address}'",
            **get_fields_from_profile(profile),
        )

    def delete(self, address: str) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ProfilesRepositoryInterface.delete.__doc__
        )

        self.client.delete(TABLE_NAME, address=address)

    def delete_all(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ProfilesRepositoryInterface.delete_all.__doc__
        )

        self.client.clear(TABLE_NAME)

    def exists(self, address: str) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ProfilesRepositoryInterface.exists.__doc__
        )

        row = self.client.select_one(
            f"SELECT address FROM {TABLE_NAME} WHERE address=?", (address,)
        )
        if row is None:
            return False

        return True


def get_fields_from_profile(profile: Profile) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param profile: Profile instance
    :return:
    """
    fields = {}
    for (key, value) in profile.__dict__.items():
        if key.startswith("_"):
            continue
        if key == "data":
            # convert dict into json string
            value = json.dumps(value)
        fields[key] = value

    return fields


def get_profile_from_row(row: tuple) -> Profile:
    """
    Return a Profile instance from a result set row

    :param row: Result set row
    :return:
    """
    values: List[Any] = []
    count = 0
    for value in row:
        if count == 1:
            # convert json string into dict
            values.append(json.loads(value))
        else:
            values.append(value)
        count += 1

    return Profile(*values)  # pylint: disable=no-value-for-parameter



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass
from typing import Any, OrderedDict


@dataclass
class Profile:
    address: str
    data: OrderedDict[str, Any]



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, Optional
from uuid import UUID, uuid4

from tikka.domains.accounts import Accounts
from tikka.domains.entities.category import Category
from tikka.domains.entities.events import CategoryEvent
from tikka.domains.events import EventDispatcher
from tikka.interfaces.adapters.repository.accounts import AccountsRepositoryInterface
from tikka.interfaces.adapters.repository.categories import (
    CategoriesRepositoryInterface,
)


class Categories:

    """
    Categories domain class
    """

    def __init__(
        self,
        repository: CategoriesRepositoryInterface,
        accounts: Accounts,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init Categories domain

        :param repository: AccountsRepositoryInterface instance
        :param accounts: Accounts domain instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.repository = repository
        self.accounts = accounts
        self.event_dispatcher = event_dispatcher

    @staticmethod
    def create(name: str, parent_id: Optional[UUID] = None):
        """
        Return a new Category instance with a unique ID

        :param name: Name of the category
        :param parent_id: UUID of the parent category
        :return:
        """
        return Category(id=uuid4(), name=name, parent_id=parent_id)

    def add(self, category: Category):
        """
        Add category in repository

        :param category: Category instance
        :return:
        """
        self.repository.add(category)
        self.event_dispatcher.dispatch_event(
            CategoryEvent(CategoryEvent.EVENT_TYPE_ADD, category)
        )

    def update(self, category: Category):
        """
        Update category in repository

        :param category: Category instance
        :return:
        """
        self.repository.update(category)
        self.event_dispatcher.dispatch_event(
            CategoryEvent(CategoryEvent.EVENT_TYPE_UPDATE, category)
        )

    def get(self, id: UUID) -> Optional[Category]:  # pylint: disable=redefined-builtin
        """
        Get category instance

        :param id: Category ID
        :return:
        """
        return self.repository.get(id)

    def delete(self, id: UUID) -> None:
        """
        Delete category and all its descendants recursively

        :param id: Category ID to delete
        :return: None
        """
        category = self.get(id)
        if category is None:
            return None

        # Récupérer tous les enfants directs
        children = self.list(id)

        # Supprimer récursivement tous les enfants
        for child in children:
            self.delete(child.id)

        # Supprimer les comptes de cette catégorie
        accounts = self.accounts.get_list(
            filters={AccountsRepositoryInterface.COLUMN_CATEGORY_ID: id.hex}
        )
        for account in accounts:
            self.accounts.delete(account)

        # Supprimer cette catégorie
        self.repository.delete(id)

        return None

    def delete_all(self) -> None:
        """
        Delete all categories in repository

        :return:
        """
        self.repository.delete_all()

    def list(self, parent_id: Optional[UUID]) -> List[Category]:
        """
        Return list of Category instances with parent_id

        :param parent_id: Parent ID
        :return:
        """
        return self.repository.list(parent_id)

    def list_all(self) -> List[Category]:
        """
        Return list with all Category instances

        :return:
        """
        return self.repository.list_all()

    def expand(self, category: Category):
        """
        Update category with expanded attribute to True

        :param category: Category instance
        :return:
        """
        category.expanded = True
        self.repository.update(category)

    def collapse(self, category: Category):
        """
        Update category with expanded attribute to False

        :param category: Category instance
        :return:
        """
        category.expanded = False
        self.repository.update(category)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QModelIndex, QPoint, QRect, Qt
from PyQt5.QtGui import QBrush, QColor, QFontMetrics, QPainter, QPixmap
from PyQt5.QtWidgets import (
    QApplication,
    QMainWindow,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QWidget,
)

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.category import Category
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.domains.entities.events import (
    CurrencyEvent,
    UnitEvent,
)
from tikka.slots.pyqt.entities.constants import (
    NUMERIC_DISPLAY_COLOR_BLUE,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.worker import ThreadManager
from tikka.slots.pyqt.models.account_tree import (
    AccountTreeModel,
    BalanceStyledRole,
    WalletLockStatusRole,
)
from tikka.slots.pyqt.resources.gui.widgets.account_tree_rc import Ui_AccountTreeWidget
from tikka.slots.pyqt.widgets.account_menu import AccountPopupMenu
from tikka.slots.pyqt.widgets.category_menu import CategoryPopupMenu


class AccountIconsItemDelegate(QStyledItemDelegate):
    """
    Class used to display member status icon for accounts
    """

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        """
        Draw member status icon and wallet lock status icon side by side

        :param painter: QPainter instance
        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        if not index.isValid():
            return
        super().paint(painter, option, index)

        wallet_lock_status_pixmap = index.data(WalletLockStatusRole)
        if isinstance(wallet_lock_status_pixmap, QPixmap):
            icon_x = option.rect.x() - wallet_lock_status_pixmap.rect().width() - 2
            icon_y = (
                option.rect.y()
                + (option.rect.height() - wallet_lock_status_pixmap.height()) // 2
            )
            icon_rect = QRect(
                icon_x,
                icon_y,
                wallet_lock_status_pixmap.width(),
                wallet_lock_status_pixmap.height(),
            )
            painter.drawPixmap(icon_rect, wallet_lock_status_pixmap)


class StyledBalanceItemDelegate(QStyledItemDelegate):
    """
    Class used to display balance with a custom style
    """

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        """
        Return balance with a custom style

        :param painter: QPainter instance
        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        if not index.isValid():
            return
        super().paint(painter, option, index)

        # draw balance with style
        balance_styled: str = index.data(BalanceStyledRole)
        if balance_styled:
            painter.save()

            # Enable anti-aliasing for smoother rendering
            painter.setRenderHint(QPainter.Antialiasing)

            # Set text color and font
            painter.setPen(Qt.white)
            painter.setFont(option.font)

            # Get text size using boundingRect
            font_metrics = QFontMetrics(option.font)
            text_rect = font_metrics.boundingRect(balance_styled)

            # Define padding and right margin
            padding_x, padding_y = 8, 4
            right_margin = 10  # Adjust this value for more space on the right

            # Adjust background rectangle size to include the right margin
            bg_width = text_rect.width() + 2 * padding_x + right_margin
            bg_height = text_rect.height() + 2 * padding_y

            # Position the rectangle to be right-aligned
            rect_x = option.rect.right() - bg_width - 5  # 5px margin from the right
            rect_y = option.rect.center().y() - bg_height // 2  # Centered vertically

            # Draw rounded rectangle
            painter.setBrush(
                QBrush(QColor(NUMERIC_DISPLAY_COLOR_BLUE))
            )  # Blue background
            # painter.setPen(Qt.NoPen)
            painter.drawRoundedRect(QRect(rect_x, rect_y, bg_width, bg_height), 10, 10)

            # Draw text inside the rectangle (horizontally centered with right margin)
            painter.drawText(
                QRect(
                    rect_x + padding_x,
                    rect_y + padding_y,
                    bg_width - padding_x - right_margin,
                    bg_height - 2 * padding_y,
                ),
                Qt.AlignCenter,
                balance_styled,
            )

            painter.restore()


class AccountTreeWidget(QWidget, Ui_AccountTreeWidget):
    """
    AccountTreeWidget class
    """

    def __init__(
        self,
        application: Application,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init AccountTreeWidget instance

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.event_bus = EventBus.instance()
        self.event_bus.repository_data_updated.connect(self.on_event_bus)
        self.event_bus.category_deleted.connect(self.on_event_bus)
        self.event_bus.account_added.connect(self.on_event_bus)
        self.event_bus.account_forgotten.connect(self.on_event_bus)
        self.event_bus.account_renamed.connect(self.on_event_bus)
        self.event_bus.account_wallet_updated.connect(self.on_event_bus)
        self.event_bus.identities_updated.connect(self.on_event_bus)
        self.event_bus.account_identities_updated.connect(self.on_event_bus)
        self.event_bus.identity_certified.connect(self.on_event_bus)
        self.event_bus.account_balances_updated.connect(self.on_event_bus)

        self.error_message: Optional[str] = None

        self.account_tree_model = AccountTreeModel(self.application, self.locale())
        self.treeView.setModel(self.account_tree_model)
        self.treeView.setHeaderHidden(True)
        self.treeView.expandAll()
        self.treeView.resizeColumnToContents(0)

        # set style sheet
        self.treeView.setStyleSheet("QTreeView::item {" "padding: 5px;" "}")
        self.account_icons_item_delegate = AccountIconsItemDelegate()
        self.styled_balance_item_delegate = StyledBalanceItemDelegate()
        self.treeView.setItemDelegateForColumn(0, self.account_icons_item_delegate)
        self.treeView.setItemDelegateForColumn(3, self.styled_balance_item_delegate)

        # init expand/collapse state of categories
        self.recursive_expand_or_collapse_categories(QModelIndex())
        self.treeView.resizeColumnToContents(1)

        ##############################
        # ASYNC METHODS
        ##############################
        self.thread_manager: ThreadManager = ThreadManager(self)

        # events
        self.addCategoryButton.clicked.connect(self.on_add_category_button_clicked)
        self.refreshButton.clicked.connect(self.on_refresh_button_clicked)
        self.treeView.customContextMenuRequested.connect(self.on_context_menu)
        self.treeView.expanded.connect(self.on_expanded)
        self.treeView.collapsed.connect(self.on_collapsed)
        self.treeView.model().dataChanged.connect(self.on_data_changed)
        self.treeView.doubleClicked.connect(self.on_account_tree_double_clicked)

        # application events
        self.application.event_dispatcher.add_event_listener(
            UnitEvent.EVENT_TYPE_CHANGED, self.on_unit_event
        )
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self.on_currency_event
        )

        # display total balance
        self.update_ui()

    def on_account_tree_double_clicked(self, index: QModelIndex):
        """
        When a row is double-clicked in account tree view

        :param index: QModelIndex instance
        :return:
        """
        account = index.internalPointer().element
        if isinstance(account, Account):
            self.event_bus.account_selected.emit(account)

    def rename_category(self):
        """
        Set the current selected category in edit mode in tree view

        :return:
        """
        index = self.treeView.currentIndex()
        item = index.internalPointer()
        if isinstance(item.element, Category):
            self.treeView.edit(self.treeView.currentIndex())

    def add_category(self, parent: QModelIndex):
        """
        Add a new category in index in edit mode

        :param parent: Parent QModelIndex instance
        :return:
        """
        self.treeView.model().insertRow(0, parent)
        self.recursive_expand_or_collapse_categories(QModelIndex())
        self.treeView.edit(self.treeView.model().index(0, 0, parent))

    def add_sub_category(self):
        """
        Add new sub category in current selected category

        :return:
        """
        index = self.treeView.currentIndex()
        item = index.internalPointer()
        if isinstance(item.element, Category):
            self.add_category(index)

    def on_data_changed(self, _):
        """
        Triggered when data changed in tree view model

        :return:
        """
        self._update_model()

    def recursive_expand_or_collapse_categories(self, index: QModelIndex) -> None:
        """
        Expand or collapse categories in tree view depending on expanded attribute

        :param index: Index to expand/collapse
        :return:
        """
        item = index.internalPointer()
        if item is not None:
            element = index.internalPointer().element
            if isinstance(element, Category):
                if element.expanded is True:
                    self.treeView.expand(index)
                else:
                    self.treeView.collapse(index)
            else:
                self.treeView.expand(index)
        for row in range(0, self.account_tree_model.rowCount(index)):
            self.recursive_expand_or_collapse_categories(
                self.account_tree_model.index(row, 0, index)
            )
        self.treeView.resizeColumnToContents(0)

    def on_expanded(self, index: QModelIndex) -> None:
        """
        Triggered when user expand a row

        :param index: QModelIndex instance of expanded row
        :return:
        """
        # get selected element
        element = index.internalPointer().element
        if isinstance(element, Category) and element.expanded is False:
            self.application.categories.expand(element)

    def on_collapsed(self, index: QModelIndex) -> None:
        """
        Triggered when user collapse a row

        :param index: QModelIndex instance of collapsed row
        :return:
        """
        # get selected element
        element = index.internalPointer().element
        if isinstance(element, Category) and element.expanded is True:
            self.application.categories.collapse(element)

    def _update_model(self):
        """
        Triggered when adding a category

        :return:
        """
        self.treeView.model().init_data()
        self.treeView.resizeColumnToContents(0)
        self.recursive_expand_or_collapse_categories(QModelIndex())

    def update_ui(self) -> None:
        """
        Update UI data

        :return:
        """
        if self.error_message is None:
            self.errorLabel.setText(self._(self.error_message))
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            amount = self.application.amounts.get_amount(unit_preference)
        else:
            amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)
        total_balance = sum(
            account.balance
            for account in self.application.accounts.get_list()
            if account.balance is not None
        )
        self.totalBalanceValueLabel.setText(
            self.locale().toCurrencyString(
                amount.value(total_balance),
                amount.symbol(),
            )
        )

    def on_add_category_button_clicked(self):
        """
        Triggered when user click on addCategoryButton

        :return:
        """
        self.add_category(self.treeView.rootIndex())

    def on_refresh_button_clicked(self):
        """
        Triggered when user click on Refresh button

        :return:
        """
        self.refreshButton.setEnabled(False)
        self.errorLabel.setText("")
        self.error_message = None
        self.thread_manager.start(
            self.fetch_from_network, self.on_finished_fetch_from_network
        )

    def on_unit_event(self, _):
        """
        When a unit event is triggered

        :return:
        """
        self._update_model()
        self.update_ui()

    def on_currency_event(self, _):
        """
        When a currency event is triggered

        :return:
        """
        self._update_model()
        self.update_ui()

    def on_context_menu(self, position: QPoint):
        """
        When right button on table widget

        :param position: QPoint instance
        :return:
        """
        # get selected account
        current_element = self.treeView.currentIndex().internalPointer().element
        if isinstance(current_element, Account):
            # display account popup menu at click position
            AccountPopupMenu(self.application, current_element, self).exec_(
                self.treeView.mapToGlobal(position)
            )
        else:
            # display category popup menu at click position
            CategoryPopupMenu(self.application, current_element, self).exec_(
                self.treeView.mapToGlobal(position)
            )

    def fetch_from_network(self):
        """
        Fetch table model accounts data from the network

        :return:
        """
        accounts = self.application.accounts.get_list()
        try:
            self.application.accounts.network_update_balances(accounts)
        except Exception as exception:
            logging.exception(exception)
            self.error_message = str(exception)

        addresses = [account.address for account in accounts]
        try:
            self.application.identities.network_update_identities(addresses)
        except Exception as exception:
            logging.exception(exception)
            self.error_message = str(exception)
        else:
            identity_indice = self.application.identities.list_indice()
            try:
                self.application.smiths.network_update_smiths(identity_indice)
            except Exception as exception:
                logging.exception(exception)
                self.error_message = str(exception)

    def on_finished_fetch_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        logging.debug("Account tree widget update")

        self.refreshButton.setEnabled(True)
        self._update_model()
        self.update_ui()

    def on_event_bus(self, _=None):
        """
        Refresh UI

        :return:
        """
        self._update_model()
        self.update_ui()
        logging.debug("Account tree widget updated by event bus")


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.showMaximized()

    main_window.setCentralWidget(AccountTreeWidget(application_, main_window))

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Any, Dict, List, Optional

from tikka.domains.entities.identity import Certification
from tikka.domains.events import EventDispatcher
from tikka.interfaces.adapters.network.indexer.certifications import (
    IndexerCertificationsInterface,
)
from tikka.interfaces.adapters.repository.certifications import (
    CertificationsRepositoryInterface,
)


class Certifications:
    """
    Certifications domain class
    """

    def __init__(
        self,
        repository: CertificationsRepositoryInterface,
        indexer_certifications: IndexerCertificationsInterface,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init Certifications domain

        :param repository: TransfersRepositoryInterface adapter instance
        :param indexer_certifications: NodeTransfersInterface adapter instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.repository = repository
        self.indexer_certifications = indexer_certifications
        self.event_dispatcher = event_dispatcher

    def set_list(
        self, identity_index: int, certifications: List[Certification]
    ) -> None:
        """
        Set identity certification list in repository

        :param identity_index: Identity index
        :param certifications: List of Certification instances
        :return:
        """
        self.repository.set_list(identity_index, certifications)

    def list(
        self,
        identity_index: int,
        filters: Optional[Dict[str, Any]] = None,
        sort_column: str = CertificationsRepositoryInterface.COLUMN_EXPIRE_ON,
        sort_order: str = CertificationsRepositoryInterface.SORT_ORDER_ASCENDING,
        limit: int = 10,
        offset: int = 0,
    ) -> List[Certification]:
        """
        List certifications from and to identity_index, from repository with optional filters and sort_column

        :param identity_index: Identity_index
        :param filters: Dict with {column: value} filters or None
        :param sort_column: Sort column constant, TransfersRepositoryInterface.COLUMN_TIMESTAMP by default
        :param sort_order: Sort order constant, TransfersRepositoryInterface.SORT_ORDER_DESCENDING by default
        :param limit: Number of rows to return
        :param offset: Offset of first row to return
        :return:
        """
        return self.repository.list(
            identity_index, filters, sort_column, sort_order, limit, offset
        )

    def delete_by_address(self, address: str, exclude: List[str]) -> None:
        """
        Delete certification history issued from or received by address
        (exclude certifications with address in exclude list)

        :param address: Account address of certifications to delete
        :param exclude: List of address to not delete
        :return:
        """
        self.repository.delete_by_address(address, exclude)

    def delete_all(self) -> None:
        """
        Delete all certifications in repository

        :return:
        """
        self.repository.delete_all()

    def network_fetch_certifications_for_identity(self, identity_index: int) -> None:
        """
        Fetch certifications from and to identity_index

        :param identity_index: Identity index
        :return:
        """
        certifications = self.indexer_certifications.list(identity_index, 1000, 0)
        self.set_list(identity_index, certifications)

    def count(self, identity_index: int) -> int:
        """
        Return count of certifications issued by identity_index

        :param identity_index: Identity index
        :return:
        """
        return len(self.list(identity_index))

    def count_received(self, identity_index: int) -> int:
        """
        Return count of certifications received by identity_index

        :param identity_index: Identity index
        :return:
        """
        return len(
            self.list(
                identity_index,
                filters={
                    CertificationsRepositoryInterface.COLUMN_RECEIVER_IDENTITY_INDEX: identity_index
                },
            )
        )



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/identity_create.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_IdentityCreateDialog(object):
    def setupUi(self, IdentityCreateDialog):
        IdentityCreateDialog.setObjectName("IdentityCreateDialog")
        IdentityCreateDialog.resize(836, 195)
        self.verticalLayout = QtWidgets.QVBoxLayout(IdentityCreateDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.helpLabel = QtWidgets.QLabel(IdentityCreateDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.helpLabel.setFont(font)
        self.helpLabel.setObjectName("helpLabel")
        self.verticalLayout.addWidget(self.helpLabel)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.addressLabel = QtWidgets.QLabel(IdentityCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addressLabel.sizePolicy().hasHeightForWidth())
        self.addressLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.addressLabel.setFont(font)
        self.addressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.addressLabel.setObjectName("addressLabel")
        self.gridLayout.addWidget(self.addressLabel, 1, 0, 1, 1)
        self.accountComboBox = QtWidgets.QComboBox(IdentityCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accountComboBox.sizePolicy().hasHeightForWidth())
        self.accountComboBox.setSizePolicy(sizePolicy)
        self.accountComboBox.setObjectName("accountComboBox")
        self.gridLayout.addWidget(self.accountComboBox, 0, 1, 1, 1)
        self.addressValueLabel = QtWidgets.QLabel(IdentityCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addressValueLabel.sizePolicy().hasHeightForWidth())
        self.addressValueLabel.setSizePolicy(sizePolicy)
        self.addressValueLabel.setText("")
        self.addressValueLabel.setObjectName("addressValueLabel")
        self.gridLayout.addWidget(self.addressValueLabel, 1, 1, 1, 1)
        self.identityIndexValueLabel = QtWidgets.QLabel(IdentityCreateDialog)
        self.identityIndexValueLabel.setStyleSheet("color: blue;")
        self.identityIndexValueLabel.setText("")
        self.identityIndexValueLabel.setObjectName("identityIndexValueLabel")
        self.gridLayout.addWidget(self.identityIndexValueLabel, 2, 1, 1, 1)
        self.accountLabel = QtWidgets.QLabel(IdentityCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accountLabel.sizePolicy().hasHeightForWidth())
        self.accountLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.accountLabel.setFont(font)
        self.accountLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.accountLabel.setObjectName("accountLabel")
        self.gridLayout.addWidget(self.accountLabel, 0, 0, 1, 1)
        self.identityIndexLabel = QtWidgets.QLabel(IdentityCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.identityIndexLabel.sizePolicy().hasHeightForWidth())
        self.identityIndexLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.identityIndexLabel.setFont(font)
        self.identityIndexLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.identityIndexLabel.setObjectName("identityIndexLabel")
        self.gridLayout.addWidget(self.identityIndexLabel, 2, 0, 1, 1)
        self.createIdentityButton = QtWidgets.QPushButton(IdentityCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.createIdentityButton.sizePolicy().hasHeightForWidth())
        self.createIdentityButton.setSizePolicy(sizePolicy)
        self.createIdentityButton.setObjectName("createIdentityButton")
        self.gridLayout.addWidget(self.createIdentityButton, 0, 2, 1, 1)
        self.identityStatusLabel = QtWidgets.QLabel(IdentityCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.identityStatusLabel.setFont(font)
        self.identityStatusLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.identityStatusLabel.setObjectName("identityStatusLabel")
        self.gridLayout.addWidget(self.identityStatusLabel, 3, 0, 1, 1)
        self.identityStatusValueLabel = QtWidgets.QLabel(IdentityCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.identityStatusValueLabel.sizePolicy().hasHeightForWidth())
        self.identityStatusValueLabel.setSizePolicy(sizePolicy)
        self.identityStatusValueLabel.setText("")
        self.identityStatusValueLabel.setObjectName("identityStatusValueLabel")
        self.gridLayout.addWidget(self.identityStatusValueLabel, 3, 1, 1, 1)
        self.loaderIconLabel = QtWidgets.QLabel(IdentityCreateDialog)
        self.loaderIconLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.loaderIconLabel.setText("")
        self.loaderIconLabel.setScaledContents(True)
        self.loaderIconLabel.setObjectName("loaderIconLabel")
        self.gridLayout.addWidget(self.loaderIconLabel, 1, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.errorLabel = QtWidgets.QLabel(IdentityCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.errorLabel.sizePolicy().hasHeightForWidth())
        self.errorLabel.setSizePolicy(sizePolicy)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout.addWidget(self.errorLabel)
        self.buttonBox = QtWidgets.QDialogButtonBox(IdentityCreateDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setCenterButtons(True)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(IdentityCreateDialog)
        self.buttonBox.accepted.connect(IdentityCreateDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(IdentityCreateDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(IdentityCreateDialog)

    def retranslateUi(self, IdentityCreateDialog):
        #  _translate = QtCore.QCoreApplication.translate
        IdentityCreateDialog.setWindowTitle(_("Create Identity"))
        self.helpLabel.setText(_("Select the account of the new identity"))
        self.addressLabel.setText(_("Address"))
        self.accountLabel.setText(_("Account"))
        self.identityIndexLabel.setText(_("Identity Index"))
        self.createIdentityButton.setText(_("Create Identity"))
        self.identityStatusLabel.setText(_("Identity Status"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Optional

from tikka.domains.entities.profile import Profile
from tikka.interfaces.adapters.network.datapod.datapod import NetworkDataPodInterface
from tikka.libs.keypair import Keypair


class DataPodProfilesInterface(abc.ABC):
    """
    DataPodProfilesInterface class
    """

    def __init__(self, datapod: NetworkDataPodInterface) -> None:
        """
        Init DataPodProfilesInterface instance

        :param datapod: NetworkDataPodInterface instance
        :return:
        """
        self.datapod = datapod

    @abc.abstractmethod
    def get(self, address: str) -> Optional[Profile]:
        """
        Return Profile instance by V2 address from datapod or None

        :param address: Profile account address
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def add(self, keypair: Keypair, profile: Profile) -> None:
        """
        Add signed profile to network server

        :param keypair: Keypair instance of profile owner
        :param profile: Profile instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, keypair: Keypair) -> None:
        """
        Delete signed profile in network server

        :param keypair: Keypair instance of profile owner
        :return:
        """
        raise NotImplementedError


class DataPodProfilesException(Exception):
    """
    DataPodProfilesException class
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import random
from typing import List, Optional

from tikka.adapters.network.datapod.datapod import NetworkDataPod
from tikka.domains.config import Config
from tikka.domains.currencies import Currencies
from tikka.domains.entities.constants import (
    DATAPOD_CESIUM_PLUS_V1_BLOCK_ZERO_HASH,
    DATAPODS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY,
)
from tikka.domains.entities.datapod import DataPod
from tikka.domains.entities.events import DataPodsEvent
from tikka.domains.events import EventDispatcher
from tikka.domains.preferences import Preferences
from tikka.interfaces.adapters.network.datapod.datapod import NetworkDataPodInterface
from tikka.interfaces.adapters.repository.datapods import DataPodsRepositoryInterface


class DataPods:
    """
    DataPods domain class
    """

    CONFIG_DATAPOD_ENDPOINTS_KEYWORD = "datapod"

    def __init__(
        self,
        repository: DataPodsRepositoryInterface,
        preferences: Preferences,
        network_datapod: NetworkDataPodInterface,
        config: Config,
        currencies: Currencies,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init DataPods domain instance

        :param repository: DataPodsRepositoryInterface instance
        :param preferences: Preferences domain instance
        :param network_datapod: Network adapter instance for handling datapods
        :param config: Config instance
        :param currencies: Currencies instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.repository = repository
        self.preferences = preferences
        self.network_datapod = network_datapod
        self.config = config
        self.currencies = currencies
        self.event_dispatcher = event_dispatcher

        self._current_url = self.currencies.get_entry_point_urls()[
            self.CONFIG_DATAPOD_ENDPOINTS_KEYWORD
        ][0]

        self.init_repository()

    def init_repository(self):
        """
        Init repository with default entry points from config

        :return:
        """
        repository_urls = self.repository.get_urls()
        currency_endpoints = self.currencies.get_entry_point_urls()

        # init repository with current currency entry point urls
        for url in currency_endpoints[self.CONFIG_DATAPOD_ENDPOINTS_KEYWORD]:
            if url not in repository_urls:
                self.repository.add(DataPod(url))

        self._current_url = self.repository.list(0, 1)[0].url

        current_url_in_preferences = self.preferences.get(
            DATAPODS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY
        )
        if (
            current_url_in_preferences is None
            or current_url_in_preferences not in self.repository.get_urls()
        ):
            self.preferences.set(
                DATAPODS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY, self.get_current_url()
            )
        else:
            self._current_url = current_url_in_preferences

    def add(self, datapod: DataPod) -> None:
        """
        Add datapod in repository

        :param datapod: DataPod instance
        :return:
        """
        self.repository.add(datapod)

        self.event_dispatcher.dispatch_event(
            DataPodsEvent(DataPodsEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def get(self, url: str) -> Optional[DataPod]:
        """
        Get DataPod instance by url

        :param url: Url
        :return:
        """
        return self.repository.get(url)

    def update(self, datapod: DataPod) -> None:
        """
        Update datapod in repository

        :param datapod: DataPod instance
        :return:
        """
        self.repository.update(datapod)

        self.event_dispatcher.dispatch_event(
            DataPodsEvent(DataPodsEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def list(self) -> List[DataPod]:
        """
        Return all DataPods from repository

        :return:
        """
        return self.repository.list()

    def count(self) -> int:
        """
        Return total datapods count

        :return:
        """
        return self.repository.count()

    def delete(self, url: str) -> None:
        """
        Delete DataPod by url

        :param url: DataPod url
        :return:
        """
        # do not delete default entry points from config
        if (
            url
            in self.currencies.get_entry_point_urls()[
                self.CONFIG_DATAPOD_ENDPOINTS_KEYWORD
            ]
        ):
            return

        self.repository.delete(url)
        # switch current entry point to first in list
        self.set_current_url(self.repository.list(0, 1)[0].url)
        # set new entry point in preferences
        self.preferences.set(
            DATAPODS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY, self.get_current_url()
        )

        self.event_dispatcher.dispatch_event(
            DataPodsEvent(DataPodsEvent.EVENT_TYPE_LIST_CHANGED)
        )

    def delete_all(self) -> None:
        """
        Delete all DataPods in repository

        :return:
        """
        self.repository.delete_all()

    def get_current_url(self) -> str:
        """
        Return current entry point url

        :return:
        """
        return self._current_url

    def set_current_url(self, url: str) -> None:
        """
        Set current entry point url

        :return:
        """
        self._current_url = url
        # update preference
        self.preferences.set(
            DATAPODS_CURRENT_ENTRY_POINT_URL_PREFERENCES_KEY,
            self._current_url,
        )

    def network_fetch_current_datapod(self) -> None:
        """
        Update datapod from network

        :return:
        """
        current_datapod = self.repository.get(self.get_current_url())
        network_datapod = self.network_datapod.get()
        if network_datapod is None:
            return None

        if current_datapod is not None:
            # update only changing properties
            current_datapod.block = network_datapod.block

            self.repository.update(current_datapod)

        return None

    def network_get_genesis_hash(self) -> str:
        """
        Get from network and return datapod genesis hash

        :return:
        """
        return self.network_datapod.get_genesis_hash()

    @staticmethod
    def network_test_and_get_datapod(url: str) -> Optional[DataPod]:
        """
        Try to open connection on url and return data pod if successful

        Then close connection

        :param url: Entry point url
        :return:
        """
        datapod = None

        network_datapod = NetworkDataPod()
        network_datapod.connection.connect(DataPod(url))
        if network_datapod.connection.is_connected():
            datapod = network_datapod.get()
        network_datapod.connection.disconnect()

        return datapod

    def network_set_url_randomly(self):
        """
        Shuffle data pods list randomly and connect to first available data pod

        :return:
        """
        indices = list(range(1, self.count()))
        random.shuffle(indices)
        for index in indices:
            url = self.list()[index].url
            # never choose localhost randomly...
            if "localhost" not in url and "127.0.0.1" not in url:
                self.network_datapod.connection.connect(DataPod(url))
                datapod_genesis_hash = self.network_datapod.get_genesis_hash()
                if not self.check_genesis_hash(datapod_genesis_hash):
                    self.network_datapod.connection.disconnect()
                if self.network_datapod.connection.is_connected():
                    self.set_current_url(url)
                    break

    def connect(self):
        """
        Connection to the current datapod or a random datapod, depending on Config

        :return:
        """
        if self.config.get(self.config.RANDOM_CONNECTION_AT_START_KEY) is True:
            self.network_set_url_randomly()

        current_datapod = self.get(self.get_current_url())
        if current_datapod is not None:
            # connect to datapod
            self.network_datapod.connection.connect(current_datapod)
            if self.network_datapod.connection.is_connected():
                datapod_genesis_hash = self.network_datapod.get_genesis_hash()
                if not self.check_genesis_hash(datapod_genesis_hash):
                    self.network_datapod.connection.disconnect()
                    logging.error("DataPod currency is different! Force Disconnect!")

    def check_genesis_hash(self, hash_to_check: str) -> bool:
        """
        Return True if hash == datapod genesis hash

        :param hash_to_check: Hash to check
        :return:
        """
        # do not check if localhost
        if (
            "localhost" in self.network_datapod.connection.url
            or "127.0.0.1" in self.network_datapod.connection.url
        ):
            return True

        return hash_to_check == DATAPOD_CESIUM_PLUS_V1_BLOCK_ZERO_HASH



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QMovie
from PyQt5.QtWidgets import QApplication, QDialog, QLabel, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.address import DisplayAddress
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.identity import Identity, IdentityStatus
from tikka.slots.pyqt.entities.constants import ICON_LOADER
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.worker import (
    ThreadManager,
)
from tikka.slots.pyqt.resources.gui.windows.identity_certify_rc import (
    Ui_IdentityCertifyDialog,
)
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class IdentityCertifyWindow(QDialog, Ui_IdentityCertifyDialog):
    """
    IdentityCertifyWindow class
    """

    identity_certified_signal = pyqtSignal(Identity, Identity)

    def __init__(
        self,
        application: Application,
        issuer_account: Account,
        receiver_account: Optional[Account] = None,
        parent: Optional[QWidget] = None,
    ):
        """
        Init identity certify window

        :param application: Application instance
        :param issuer_account: Creator Account instance
        :param receiver_account: Creator Account instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.event_bus = EventBus.instance()

        self.issuer_account: Account = issuer_account
        self.issuer_identity: Optional[
            Identity
        ] = self.application.identities.get_by_address(self.issuer_account.address)

        self.receiver_account: Optional[Account] = receiver_account
        self.receiver_identity: Optional[Identity] = None

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        # events
        if self.receiver_account is None:
            self.identityComboBox.activated.connect(
                self.on_identity_combobox_index_changed
            )
            self.certifyIdentityButton.setEnabled(False)
        else:
            self.helpLabel.hide()
            self.identityComboBox.hide()
            self.receiver_identity = self.application.identities.get_by_address(
                self.receiver_account.address
            )
            if self.receiver_identity is not None:
                self.gridLayout.addWidget(QLabel(self.receiver_account.name or ""))
                self.update_ui()

        self.certifyIdentityButton.clicked.connect(
            self._on_certify_identity_button_clicked
        )

        ##############################
        # ASYNC METHODS
        ##############################
        self.thread_manager: ThreadManager = ThreadManager(self)

        self.init_identity_combo_box()

    def init_identity_combo_box(self) -> None:
        """
        Init combobox with identities not unconfirmed nor revoked
        Avec séparateur entre noms et adresses
        """
        self.identityComboBox.clear()
        self.identityComboBox.addItem("-", userData=None)

        for identity in self.application.identities.list():
            if (
                identity.status
                not in (
                    IdentityStatus.UNCONFIRMED,
                    IdentityStatus.REVOKED,
                )
                and identity.address != self.issuer_account.address
            ):
                account = self.application.accounts.get_by_address(identity.address)
                account_name = f" - {account.name}" if account.name is not None else ""  # type: ignore

                self.identityComboBox.addItem(
                    f"{identity.name}#{identity.index} - {DisplayAddress(identity.address).shorten}{account_name}",
                    userData=identity.index,
                )

    def on_identity_combobox_index_changed(self):
        """
        Triggered when identity selection is changed

        :return:
        """
        self.receiver_identity = self.application.identities.get(
            self.identityComboBox.currentData()
        )
        self.certifyIdentityButton.setEnabled(self.receiver_identity is not None)
        self.update_ui()

    def update_ui(self):
        """
        Update UI fields

        :return:
        """
        if self.receiver_identity is not None:
            display_identity_status = {
                IdentityStatus.UNCONFIRMED.value: self._("Unconfirmed"),
                IdentityStatus.UNVALIDATED.value: self._("Unvalidated"),
                IdentityStatus.MEMBER.value: self._("Member"),
                IdentityStatus.NOT_MEMBER.value: self._("Not member"),
                IdentityStatus.REVOKED.value: self._("Revoked"),
            }
            identity_status_color = {
                IdentityStatus.UNCONFIRMED.value: "orange",
                IdentityStatus.UNVALIDATED.value: "orange",
                IdentityStatus.MEMBER.value: "green",
                IdentityStatus.NOT_MEMBER.value: "red",
                IdentityStatus.REVOKED.value: "red",
            }
            identity_status_style = f"""
                                    color: white;
                                    border: 3px solid "{identity_status_color[self.receiver_identity.status.value]}";
                                    border-radius: 10%;
                                    background: "{identity_status_color[self.receiver_identity.status.value]}";
                                    """
            self.addressValueLabel.setText(self.receiver_identity.address)
            self.identityIndexValueLabel.setText(f"#{self.receiver_identity.index}")
            self.identityNameValueLabel.setText(self.receiver_identity.name)
            self.identityStatusValueLabel.setText(
                display_identity_status[self.receiver_identity.status.value]
            )
            self.identityStatusValueLabel.setStyleSheet(identity_status_style)

            self.certifyIdentityButton.setEnabled(True)
        else:
            self.addressValueLabel.setText("")
            self.identityIndexValueLabel.setText("")
            self.identityNameValueLabel.setText("")
            self.identityStatusValueLabel.setText("")
            self.certifyIdentityButton.setEnabled(False)

    def _on_certify_identity_button_clicked(self):
        """
        Run when use click certify identity button

        :return:
        """
        if not self.application.wallets.is_unlocked(self.issuer_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.issuer_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.loaderIconLabel.show()
        self.certifyIdentityButton.setDisabled(True)
        self.thread_manager.start(
            self.network_certify_identity, self._on_finished_network_certify_identity
        )

    def network_certify_identity(self):
        """
        Send certify identity request to network

        :return:
        """
        if self.receiver_identity is not None:
            currency = self.application.currencies.get_current()
            try:
                assert self.issuer_identity is not None
                self.application.identities.network_certify(
                    self.application.wallets.get_keypair(self.issuer_account.address),
                    self.issuer_identity.index,
                    self.receiver_identity.index,
                )
            except Exception as exception:
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(str(exception))
                logging.exception(exception)
            else:
                self.errorLabel.setStyleSheet("color: green;")
                if (
                    self.receiver_identity.status == IdentityStatus.UNVALIDATED
                    and self.application.certifications.count_received(
                        self.receiver_identity.index
                    )
                    + 1
                    >= currency.certification_number_to_be_member
                ):
                    self.errorLabel.setText(
                        self._("Certification sent!\nRequest membership validation.")
                    )
                else:
                    self.errorLabel.setText(self._("Certification sent!"))

            if (
                self.receiver_identity.status == IdentityStatus.UNVALIDATED
                and self.application.certifications.count_received(
                    self.receiver_identity.index
                )
                + 1
                >= currency.certification_number_to_be_member
            ):
                try:
                    self.application.identities.network_request_distance_evaluation_for(
                        self.application.wallets.get_keypair(
                            self.issuer_account.address
                        ),
                        self.receiver_identity.index,
                    )
                except Exception as exception:
                    self.errorLabel.setStyleSheet("color: red;")
                    self.errorLabel.setText(str(exception))
                    logging.exception(exception)

    def _on_finished_network_certify_identity(self):
        """
        When async network_certify_identity is finished

        :return:
        """
        self.receiver_identity = self.application.identities.get(
            self.receiver_identity.index
        )

        self.certifyIdentityButton.setDisabled(False)
        self.loaderIconLabel.hide()
        self.update_ui()
        if self.receiver_identity is not None:
            self.certifyIdentityButton.setEnabled(False)
            if self.issuer_identity is not None:
                logging.debug(
                    "Identity certify window emit signal with issuer %s -> receiver %s",
                    self.issuer_identity.name,
                    self.receiver_identity.name,
                )
                self.event_bus.identity_certified.emit(
                    self.issuer_identity, self.receiver_identity
                )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    account_ = Account("xxxx")
    account2_ = Account("yyyy")

    IdentityCertifyWindow(application_, account_, account2_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import base64
import json
from collections import OrderedDict
from datetime import datetime

# import re
#
# # from copy import copy
from hashlib import sha256
from typing import Optional

import base58

from tikka.adapters.network.datapod.connection import DataPodConnection
from tikka.domains.entities.profile import Profile
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.datapod.profiles import (
    DataPodProfilesException,
    DataPodProfilesInterface,
)
from tikka.libs.keypair import Keypair


class DataPodProfiles(DataPodProfilesInterface):
    """
    DataPodProfiles class
    """

    def get(self, address: str) -> Optional[Profile]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            DataPodProfilesInterface.get.__doc__
        )
        connection: DataPodConnection = self.datapod.connection
        if not connection.is_connected() or connection.client is None:
            raise DataPodProfilesException(NetworkConnectionError())

        # for now, we use cesium+ V1 datapods
        # the datapod validate signature from v1 address (base58 public key)
        # so we need to convert V2 address to V1 address
        keypair = Keypair(address)
        v1_address = base58.b58encode(keypair.public_key).decode()

        try:
            result_string = connection.execute_query(
                endpoint=f"/user/profile/{v1_address}/_source"
            )
        except Exception as exception:
            if "HTTP 404:" in exception.args[0]:
                return None

            raise DataPodProfilesException(exception)
        #
        # {
        #    'title': 'Jean Dupond',
        #    'description': 'boulanger',
        #    'socials': [
        #       {
        #           'type': 'web',
        #           'url': 'http://jean.boulanger.fr'
        #        }
        #     ],
        #     'avatar': {
        #       '_content_type': 'image/png',
        #       '_content': 'iVBORw0KGgoAAAANSUhEUgAAAMgAAADICAYAAACtWK6eAAAgAuOuLp3zkbkD/w8GeTbwGV2uCwAAAABJRU5ErkJggg=='
        #     },
        #     'time': 1584271238,
        #     'issuer': 'xxx',
        #     'hash': 'yyy',
        #     'signature': 'xYPDg==',
        #     'city': 'Paris, 75020',
        #     'geoPoint': {
        #       'lon': 1.0,
        #       'lat': 48.0
        #     },
        #     'version': 2
        #  }
        #

        #        cleaned_json = serialize_document_for_hash(result_string)
        # # check hash
        # # Regex pour supprimer "hash": "xxx" (avec différents formats d'espaces)
        # pattern1 = r"\"hash\"\s*:\s*\"[^\"]*\"\s*,?"
        # # Regex pour supprimer "signature": "xxx"
        # pattern2 = r"\"signature\"\s*:\s*\"[^\"]*\"\s*,?"
        #
        # # Supprimer les champs sensibles
        # cleaned_json = re.sub(pattern1, "", result_string, flags=re.MULTILINE)
        # cleaned_json = re.sub(pattern2, "", cleaned_json, flags=re.MULTILINE)
        #
        # # Nettoyer les virgules orphelines (qui pourraient casser le JSON)
        # cleaned_json = re.sub(r",\s*}", "}", cleaned_json)  # Virgule avant }
        # cleaned_json = re.sub(r",\s*,", ",", cleaned_json)  # Double virgule

        # check_hash = sha256(cleaned_json.encode()).hexdigest()
        # check_hash_deepseek = calculate_hash_java_compatible(result_string)

        result = json.loads(result_string, object_pairs_hook=OrderedDict)
        # if check_hash != result["hash"].lower():
        #     return None

        # check signature

        return Profile(address=address, data=result)

    def add(self, keypair: Keypair, profile: Profile) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            DataPodProfilesInterface.add.__doc__
        )
        connection: DataPodConnection = self.datapod.connection
        if not connection.is_connected() or connection.client is None:
            raise DataPodProfilesException(NetworkConnectionError())

        v1_address = base58.b58encode(keypair.public_key).decode()

        document = profile.data
        # remove old hash and signature
        if "hash" in document:
            del document["hash"]
        if "signature" in document:
            del document["signature"]

        document["time"] = int(datetime.now().timestamp())
        document["issuer"] = v1_address
        hash = sha256(json.dumps(document).encode()).hexdigest().upper()
        signature = keypair.sign(data=hash)
        signature_base64 = base64.b64encode(signature).decode("UTF8")
        document["hash"] = hash
        document["signature"] = signature_base64
        # {
        #     "version" : 2,
        #     "title" : "Pecquot Ludovic",
        #     "description" : "Développeur Java et techno client-serveur\nParticipation aux #RML7, #EIS et #Sou",
        #     "time" : 1488359903,
        #     "issuer" : "2v6tXNxGC1BWaJtUFyPJ1wJ8rbz9v1ZVU1E1LEV2v4ss",
        #     "hash" : "F66D43ECD4D38785F424ADB68B3EA13DD56DABDE275BBE780E81E8D4E1D0C5FA",
        #     "signature" : "3CWxdLtyY8dky97RZBFLfP6axnfW8KUmhlkiaXC7BN98yg6xE9CkijRBGmuyrx3llPx5HeoGLG99DyvVIKZuCg=="
        # }
        try:
            connection.execute_query(
                endpoint="/user/profile", method="POST", data=document
            )
        except Exception as exception:
            raise DataPodProfilesException(exception)

    def delete(self, keypair: Keypair) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            DataPodProfilesInterface.delete.__doc__
        )
        connection: DataPodConnection = self.datapod.connection
        if not connection.is_connected() or connection.client is None:
            raise DataPodProfilesException(NetworkConnectionError())

        v1_address = base58.b58encode(keypair.public_key).decode()

        document = {
            "version": 2,
            "index": "user",
            "type": "profile",
            "id": v1_address,
            "issuer": v1_address,
            "time": int(datetime.now().timestamp()),
        }
        hash = sha256(json.dumps(document).encode()).hexdigest().upper()
        signature = keypair.sign(data=hash)
        signature_base64 = base64.b64encode(signature).decode("UTF8")
        document["hash"] = hash
        document["signature"] = signature_base64
        print(document)
        try:
            connection.execute_query(
                endpoint="/history/delete",
                method="POST",
                data=document
                # endpoint="/user/profile/_delete", method="POST", data=document
            )
        except Exception as exception:
            raise DataPodProfilesException(exception)


def calculate_hash_java_compatible(json_string):
    """
    Calcule le hash de la même manière que le code Java
    en conservant l'ordre des champs
    """
    # Parser le JSON en OrderedDict pour conserver l'ordre
    data = json.loads(json_string, object_pairs_hook=OrderedDict)

    # Créer un nouveau OrderedDict sans les champs sensibles
    data_for_hash = OrderedDict()
    for key, value in data.items():
        if key not in ["hash", "signature", "read_signature"]:
            data_for_hash[key] = value

    # Resérialiser exactement comme le fait Jackson
    json_for_hash = json.dumps(
        data_for_hash,
        separators=(",", ":"),  # Pas d'espaces, comme Jackson
        ensure_ascii=False
        # L'ordre est conservé par l'OrderedDict
    )

    return sha256(json_for_hash.encode()).hexdigest()


def serialize_document_for_hash(data: str) -> str:
    """
    Serialise Cesium+ document to get a corrct hash...

    :param data: Source Json Document as string
    :return:
    """
    serialized_json = OrderedDict(
        {
            "title": "",
            "description": "",
            "address": "",
            "city": "",
            "socials": [],
            "time": 0,
            "issuer": "",
            "version": 2,
            "tags": [],
            "geoPoint": {"lon": 0, "lat": 0},
            "avatar": {"_content_type": "", "_content": ""},
        }
    )

    for key, value in json.loads(data).items():
        if key not in ("hash", "signature", "read_signature"):
            serialized_json[key] = value

    return json.dumps(serialized_json, ensure_ascii=False, separators=(",", ":"))



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Dict, List, Optional

from tikka.domains.entities.identity import Identity, IdentityStatus
from tikka.interfaces.adapters.network.indexer.identities import (
    IndexerIdentitiesInterface,
)
from tikka.interfaces.adapters.network.node.identities import NodeIdentitiesInterface
from tikka.interfaces.adapters.repository.identities import (
    IdentitiesRepositoryInterface,
)
from tikka.libs.keypair import Keypair


class Identities:

    """
    Identities domain class
    """

    def __init__(
        self,
        repository: IdentitiesRepositoryInterface,
        node_identities: NodeIdentitiesInterface,
        indexer_identities: IndexerIdentitiesInterface,
    ):
        """
        Init Identities domain

        :param repository: IdentitiesRepositoryInterface instance
        :param node_identities: NodeIdentitiesInterface instance
        :param indexer_identities: IndexerIdentitiesInterface instance
        """
        self.repository = repository
        self.node_identities = node_identities
        self.indexer_identities = indexer_identities

    @staticmethod
    def create(
        index: int,
        removable_on: int,
        next_creatable_on: int,
        address: str,
        old_address: Optional[str] = None,
        status: IdentityStatus = IdentityStatus.UNCONFIRMED,
        first_eligible_ud: int = 0,
    ):
        """
        Return an identity instance from params

        :param index: Index number in blockchain
        :param removable_on: Identity expiration timestamp
        :param next_creatable_on: Date after which a new identity can be created
        :param address: Account address
        :param old_address: Previous account address
        :param status: Identity status
        :param first_eligible_ud: First elligible UD index
        :return:
        """
        return Identity(
            index=index,
            name=None,
            removable_on=removable_on,
            next_creatable_on=next_creatable_on,
            status=status,
            address=address,
            old_address=old_address,
            first_eligible_ud=first_eligible_ud,
        )

    def add(self, identity: Identity):
        """
        Add identity in repository

        :param identity: Identity instance
        :return:
        """
        self.repository.add(identity)

    def update(self, identity: Identity):
        """
        Update identity in repository

        :param identity: Identity instance
        :return:
        """
        self.repository.update(identity)

    def get(self, index: int) -> Optional[Identity]:
        """
        Get identity instance

        :param index: Identity index
        :return:
        """
        return self.repository.get(index)

    def get_by_address(self, address: str) -> Optional[Identity]:
        """
        Return Identity instance from account address or None

        :param address: Account address
        :return:
        """
        return self.repository.get_by_address(address)

    def get_by_addresses(self, addresses: List[str]) -> Dict[str, Optional[Identity]]:
        """
        Return Identity instance from account address or None

        :param addresses: List of Account addresses
        :return:
        """
        return self.repository.get_by_addresses(addresses)

    def get_index_by_address(self, address: str) -> Optional[int]:
        """
        Return identity index from account address or None

        :param address: Account address
        :return:
        """
        return self.repository.get_index_by_address(address)

    def delete(self, index: int) -> None:
        """
        Delete identity in repository

        :param index: Identity index to delete
        :return:
        """
        self.repository.delete(index)

    def delete_all(self) -> None:
        """
        Delete all identities in repository

        :return:
        """
        self.repository.delete_all()

    def exists(self, index: int) -> bool:
        """
        Return True if identity exists in repository

        :param index: Identity index to check
        :return:
        """
        return self.repository.exists(index)

    def list(self):
        """
        Return list of all identities

        :return:
        """
        return self.repository.list()

    def list_indice(self):
        """
        Return list of all identity indice

        :return:
        """
        return self.repository.list_indice()

    def is_validated(self, index: int) -> bool:
        """
        Return True if identity status is validated

        :param index: Identity index to check
        :return:
        """
        identity = self.get(index)
        if identity is None:
            return False
        return identity.status == IdentityStatus.MEMBER

    def network_update_identity(self, address: str) -> Optional[Identity]:
        """
        Update and return Identity instance by address from network if any

        :param address: Owner account address
        :return:
        """
        identity = self.node_identities.get_identity(address)
        if identity is None:
            old_identity = self.get_by_address(address)
            if old_identity is not None:
                self.delete(old_identity.index)
        else:
            if self.indexer_identities.indexer.connection.is_connected():
                identity.name = self.indexer_identities.get_identity_name(
                    identity.index
                )
            stored_identity = self.get(identity.index)
            if stored_identity is not None:
                if (
                    identity.name is None
                    and stored_identity.name is not None
                    and stored_identity.index == identity.index
                ):
                    identity.name = stored_identity.name
                self.update(identity)
            else:
                self.add(identity)

        return identity

    def network_get_identity(self, address: str) -> Optional[Identity]:
        """
        Return Identity instance by address from network if any

        :param address: Owner account address
        :return:
        """
        identity = self.node_identities.get_identity(address)

        if identity and self.indexer_identities.indexer.connection.is_connected():
            identity.name = self.indexer_identities.get_identity_name(identity.index)

        return identity

    def network_update_identities(self, addresses: List[str]) -> None:
        """
        Update repository Identities by account address list from network

        :param addresses: Account address list
        :return:
        """
        identities = self.network_get_identities(addresses)
        names: Dict[int, str] = {}
        if self.indexer_identities.indexer.connection.is_connected():
            names = self.indexer_identities.get_identity_names(
                [
                    identity.index
                    for identity in identities.values()
                    if identity is not None
                ]
            )
        for address, identity in identities.items():
            if identity is None:
                old_identity = self.get_by_address(address)
                if old_identity is not None:
                    self.delete(old_identity.index)
            else:
                identity.name = names.get(identity.index, None)
                stored_identity = self.get(identity.index)
                if stored_identity is not None:
                    if (
                        identity.name is None
                        and stored_identity.name is not None
                        and stored_identity.index == identity.index
                    ):
                        identity.name = stored_identity.name
                    self.update(identity)
                else:
                    self.add(identity)

    def network_get_identities(
        self, addresses: List[str]
    ) -> Dict[str, Optional[Identity]]:
        """
        Return Identity instances by account address list from network

        :param addresses: Account address list
        :return:
        """
        identities = self.node_identities.get_identities(addresses)
        indice = [
            identity.index for identity in identities.values() if identity is not None
        ]
        if self.indexer_identities.indexer.connection.is_connected():
            names = self.indexer_identities.get_identity_names(indice)
            for identity in identities.values():
                if identity is None:
                    continue
                identity.name = names[identity.index]

        return identities

    def network_change_owner_key(
        self, old_keypair: Keypair, new_keypair: Keypair
    ) -> None:
        """
        Change identity owner from old_keypair to new_keypair on blockchain

        :param old_keypair: Keypair of current identity account
        :param new_keypair: Keypair of new identity account
        :return:
        """
        return self.node_identities.change_owner_key(old_keypair, new_keypair)

    def network_claim_uds(self, keypair: Keypair) -> None:
        """
        Add unclaimed UDs of identity to keypair account balance

        :param keypair: Keypair of account
        :return:
        """
        return self.node_identities.claim_uds(keypair)

    def network_create(self, keypair: Keypair, address: str) -> None:
        """
        Create identity for account address

        :param keypair: Keypair of creator
        :param address: Identity account address
        :return:
        """
        return self.node_identities.create(keypair, address)

    def network_confirm(self, keypair: Keypair, name: str) -> None:
        """
        Confirm identity with a name

        :param keypair: Keypair of identity owner
        :param name: Identity name
        :return:
        """
        return self.node_identities.confirm(keypair, name)

    def network_certify(
        self, keypair: Keypair, issuer_index: int, receiver_index: int
    ) -> None:
        """
        Certify identity by index

        :param keypair: Keypair of certifier
        :param issuer_index: Issuer Identity index
        :param receiver_index: Receiver Identity index
        :return:
        """
        return self.node_identities.certify(keypair, issuer_index, receiver_index)

    def network_request_distance_evaluation_for(
        self, keypair: Keypair, identity_index: int
    ) -> None:
        """
        Request distance evaluation for identity index,
        identity status must be unvalidated

        :param keypair: Keypair of issuer
        :param identity_index: Target Identity index
        :return:
        """
        return self.node_identities.request_distance_evaluation_for(
            keypair, identity_index
        )



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/account_certification_row.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AccountCertificationRowWidget(object):
    def setupUi(self, AccountCertificationRowWidget):
        AccountCertificationRowWidget.setObjectName("AccountCertificationRowWidget")
        AccountCertificationRowWidget.resize(490, 126)
        self.verticalLayout = QtWidgets.QVBoxLayout(AccountCertificationRowWidget)
        self.verticalLayout.setContentsMargins(-1, 0, -1, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setVerticalSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.accountAddressValuelabel = QtWidgets.QLabel(AccountCertificationRowWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accountAddressValuelabel.sizePolicy().hasHeightForWidth())
        self.accountAddressValuelabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Courier")
        self.accountAddressValuelabel.setFont(font)
        self.accountAddressValuelabel.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.accountAddressValuelabel.setObjectName("accountAddressValuelabel")
        self.horizontalLayout_2.addWidget(self.accountAddressValuelabel)
        self.identityNameValueLabel = QtWidgets.QLabel(AccountCertificationRowWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.identityNameValueLabel.sizePolicy().hasHeightForWidth())
        self.identityNameValueLabel.setSizePolicy(sizePolicy)
        self.identityNameValueLabel.setIndent(10)
        self.identityNameValueLabel.setObjectName("identityNameValueLabel")
        self.horizontalLayout_2.addWidget(self.identityNameValueLabel)
        self.accountNameValueLabel = QtWidgets.QLabel(AccountCertificationRowWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accountNameValueLabel.sizePolicy().hasHeightForWidth())
        self.accountNameValueLabel.setSizePolicy(sizePolicy)
        self.accountNameValueLabel.setObjectName("accountNameValueLabel")
        self.horizontalLayout_2.addWidget(self.accountNameValueLabel)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.identityStatusValueLabel = QtWidgets.QLabel(AccountCertificationRowWidget)
        self.identityStatusValueLabel.setObjectName("identityStatusValueLabel")
        self.verticalLayout_2.addWidget(self.identityStatusValueLabel)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.createdOnValueLabel = QtWidgets.QLabel(AccountCertificationRowWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.createdOnValueLabel.sizePolicy().hasHeightForWidth())
        self.createdOnValueLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setItalic(True)
        self.createdOnValueLabel.setFont(font)
        self.createdOnValueLabel.setStyleSheet("color: \"gray\"")
        self.createdOnValueLabel.setObjectName("createdOnValueLabel")
        self.horizontalLayout.addWidget(self.createdOnValueLabel)
        self.updatedOnValueLabel = QtWidgets.QLabel(AccountCertificationRowWidget)
        font = QtGui.QFont()
        font.setItalic(True)
        self.updatedOnValueLabel.setFont(font)
        self.updatedOnValueLabel.setStyleSheet("color: \"gray\"")
        self.updatedOnValueLabel.setObjectName("updatedOnValueLabel")
        self.horizontalLayout.addWidget(self.updatedOnValueLabel)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.gridLayout.addLayout(self.verticalLayout_2, 0, 1, 1, 1)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.expireOnValueLabel = QtWidgets.QLabel(AccountCertificationRowWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.expireOnValueLabel.sizePolicy().hasHeightForWidth())
        self.expireOnValueLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.expireOnValueLabel.setFont(font)
        self.expireOnValueLabel.setStyleSheet("color: white;\n"
"border : 3px solid green;\n"
"border-radius: 10%;\n"
"background: green;")
        self.expireOnValueLabel.setAlignment(QtCore.Qt.AlignHCenter|QtCore.Qt.AlignTop)
        self.expireOnValueLabel.setObjectName("expireOnValueLabel")
        self.verticalLayout_3.addWidget(self.expireOnValueLabel, 0, QtCore.Qt.AlignBottom)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem)
        self.gridLayout.addLayout(self.verticalLayout_3, 0, 4, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.line = QtWidgets.QFrame(AccountCertificationRowWidget)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)

        self.retranslateUi(AccountCertificationRowWidget)
        QtCore.QMetaObject.connectSlotsByName(AccountCertificationRowWidget)

    def retranslateUi(self, AccountCertificationRowWidget):
        #  _translate = QtCore.QCoreApplication.translate
        AccountCertificationRowWidget.setWindowTitle(_("CertificationRowWidget"))
        self.accountAddressValuelabel.setText(_("address"))
        self.identityNameValueLabel.setText(_("name#0000"))
        self.accountNameValueLabel.setText(_(" - Name"))
        self.identityStatusValueLabel.setText(_("status"))
        self.createdOnValueLabel.setText(_("datetime"))
        self.updatedOnValueLabel.setText(_("(datetime)"))
        self.expireOnValueLabel.setText(_("datetime"))



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import logging
from typing import Any, Dict, Optional

import urllib3

from tikka.interfaces.adapters.network.connection import ConnectionInterface
from tikka.interfaces.entities.server import ServerInterface


class DataPodConnection(ConnectionInterface):
    """
    DataPodConnection class with urllib3 REST client
    """

    client: Optional[urllib3.PoolManager] = None

    def __init__(self) -> None:
        """
        Init DataPodConnection instance
        """
        self.client: Optional[urllib3.PoolManager] = None

    def connect(self, server: ServerInterface) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.connect.__doc__
        )
        self.url = server.url.rstrip("/")  # Nettoyer l'URL
        logging.debug(
            "CONNECTING TO DATAPOD %s......................................", self.url
        )

        # Créer le client HTTP
        self.client = urllib3.PoolManager(
            timeout=urllib3.Timeout(connect=5.0, read=10.0), retries=urllib3.Retry(3)
        )

        try:
            # Test de connexion avec une requête REST
            endpoint = f"{self.url}/node/summary"

            response = self.client.request("GET", endpoint)

            if response.status != 200:
                raise Exception(f"HTTP {response.status}: {response.data.decode()}")

            result = json.loads(response.data.decode())

            # Vérifier la structure de la réponse
            if (
                "duniter" not in result
                or "status" not in result["duniter"]
                or result["duniter"]["status"] != 200
            ):
                self.client = None
                logging.debug(
                    "CONNECTION TO DATAPOD %s......................................FAILED !",
                    self.url,
                )
            else:
                logging.debug(
                    "CONNECTION TO DATAPOD %s......................................SUCCESS !",
                    self.url,
                )

        except Exception as exception:
            logging.exception(exception)
            self.client = None
            logging.debug(
                "CONNECTION TO DATAPOD %s......................................FAILED !",
                self.url,
            )

    def disconnect(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.disconnect.__doc__
        )
        if self.client is not None:
            self.client.clear()
            self.client = None
            logging.debug("REST connection closed.")

    def is_connected(self) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.is_connected.__doc__
        )
        return self.client is not None

    def execute_query(
        self,
        endpoint: str,
        method: str = "GET",
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> str:
        """
        Exécute une requête REST

        :param endpoint: Point d'accès (ex: "/blocks")
        :param method: Méthode HTTP (GET, POST, etc.)
        :param params: Paramètres de query string
        :param data: Données pour les requêtes POST
        :return: Réponse JSON décodée
        """
        if self.client is None or not self.is_connected():
            raise Exception("Not connected to server")

        url = f"{self.url}/{endpoint.lstrip('/')}"

        # Préparer les paramètres
        if params:
            from urllib.parse import urlencode

            url += "?" + urlencode(params)

        # Préparer le body pour POST
        body = None
        headers = {}
        if data and method in ["POST", "PUT", "PATCH"]:
            body = json.dumps(data).encode("utf-8")
            headers = {"Content-Type": "application/json"}

        try:
            response = self.client.request(method, url, body=body, headers=headers)

            if response.status != 200:
                raise Exception(f"HTTP {response.status}: {response.data.decode()}")

            return response.data.decode()

        except Exception as e:
            logging.error(f"Query failed: {e}")
            raise



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import sys
from typing import Optional

from PyQt5.QtWidgets import QApplication, QMenu, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.identity import Certification, IdentityStatus
from tikka.slots.pyqt.windows.identity_certify import IdentityCertifyWindow
from tikka.slots.pyqt.windows.transfer import TransferWindow


class AccountCertificationsPopupMenu(QMenu):
    """
    AccountCertificationsPopupMenu class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        certification: Certification,
        parent: Optional[QWidget] = None,
    ):
        """
        Init TransferHistoryPopupMenu instance

        :param application: Application instance
        :param account: Account instance
        :param certification: Certification instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)

        self.application = application
        self.account = account
        self.certification = certification
        self._ = self.application.translator.gettext

        self.contact_address = (
            self.certification.issuer_address
            if self.certification.issuer_address != self.account.address
            else self.certification.receiver_address
        )
        self.contact_identity_name = (
            self.certification.issuer_identity_name
            if self.certification.issuer_address != self.account.address
            else self.certification.receiver_identity_name
        )
        self.contact_identity_index = (
            self.certification.issuer_identity_index
            if self.certification.issuer_address != self.account.address
            else self.certification.receiver_identity_index
        )

        # Create all menu actions and disable them based on conditions
        transfer_to_action = self.addAction(self._("Transfer to"))
        transfer_to_action.triggered.connect(self.transfer_to)

        transfer_from_action = self.addAction(self._("Transfer from"))
        transfer_from_action.triggered.connect(self.transfer_from)

        certify_action = self.addAction(self._("Certify"))
        certify_action.triggered.connect(self.certify)

        # Enable/disable actions based on conditions
        if not self.application.wallets.exists(self.account.address):
            transfer_to_action.setEnabled(False)
            certify_action.setEnabled(False)
        else:
            # Check if identity exists for certification
            if self.application.identities.get(self.contact_identity_index) is None:
                certify_action.setEnabled(False)

        if not self.application.wallets.exists(self.contact_address):
            transfer_from_action.setEnabled(False)

        # Separator
        self.addSeparator()

        # menu actions
        copy_address_to_clipboard_action = self.addAction(
            self._("Copy address to clipboard")
        )
        copy_address_to_clipboard_action.triggered.connect(
            self.copy_address_to_clipboard
        )

        self.contact_account = self.application.accounts.get_by_address(
            self.contact_address
        )

        add_account_address_action = self.addAction(self._("Add address to accounts"))
        add_account_address_action.triggered.connect(self.add_address_to_accounts)

        # Disable "Add address to accounts" if account already exists
        if self.contact_account is not None:
            add_account_address_action.setEnabled(False)

    def copy_address_to_clipboard(self):
        """
        Copy address of selected row to clipboard

        :return:
        """
        clipboard = QApplication.clipboard()
        clipboard.setText(self.contact_address)

    def transfer_to(self):
        """
        Open transfer window with account as recipient

        :return:
        """
        if self.contact_account is None:
            self.contact_account = Account(
                self.contact_address,
                f"{self.contact_identity_name}#{self.contact_identity_index}",
            )

        TransferWindow(
            self.application,
            self.account,
            self.contact_account,
            parent=self,
        ).exec_()

    def transfer_from(self):
        """
        Open transfer window with account as sender

        :return:
        """
        TransferWindow(
            self.application,
            self.contact_account,
            self.account,
            parent=self,
        ).exec_()

    def add_address_to_accounts(self):
        """
        Add a new account from contact address

        :return:
        """
        self.contact_account = Account(self.contact_address)

        self.application.accounts.add(self.contact_account)

    def certify(self):
        """
        Open certify window with account as issuer and contact_account as receiver

        :return:
        """
        self.contact_account = Account(self.contact_address)
        IdentityCertifyWindow(
            self.application,
            self.account,
            self.contact_account,
            parent=self,
        ).exec_()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    account_ = Account("732SSfuwjB7jkt9th1zerGhphs6nknaCBCTozxUcPWPU")
    certification_ = Certification(
        id="test_certification_1_2",
        issuer_identity_index=1,
        issuer_identity_name="Alice",
        issuer_identity_status=IdentityStatus.MEMBER,
        issuer_address="xxxxx",
        receiver_identity_index=2,
        receiver_identity_name="Bob",
        receiver_identity_status=IdentityStatus.MEMBER,
        receiver_address="yyyy",
        created_on=datetime.datetime.now(),
        updated_on=datetime.datetime.now(),
        expire_on=datetime.datetime.now(),
    )
    menu = AccountCertificationsPopupMenu(application_, account_, certification_)
    menu.exec_()

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
import time
from typing import Callable, Optional, Set

from PyQt5.QtCore import (
    QObject,
    QThread,
    QTimer,
    pyqtSignal,
)
from PyQt5.QtWidgets import QApplication


class TaskWorker(QObject):
    """
    TaskWorker class
    """

    finished = pyqtSignal()

    def __init__(self, call: Callable) -> None:
        """
        Init TaskWorker instance

        :param call: Callable function
        """
        super().__init__()
        self.call = call

    def run(self):
        """
        Run call

        :return:
        """
        try:
            logging.debug(
                "Worker %s running in thread: %s",
                self.call.__name__,
                QThread.currentThread(),
            )
            self.call()
        except Exception as e:
            logging.exception("Worker %s failed: %s", self.call.__name__, str(e))
        self.finished.emit()


class ThreadManager(QObject):
    """
    ThreadManager2 class
    """

    finished = pyqtSignal()

    def __init__(self, parent: Optional[QObject] = None):
        """
        Init ThreadManager2 instance

        :param parent: Parent QObject instance
        """
        super().__init__(parent)
        self._active_workers: Set = set()

    def start(
        self, task_call: Callable, finished_call: Optional[Callable] = None
    ) -> TaskWorker:
        """
        Create a worker and run it in a separate thread

        :param task_call: Task function to run
        :param finished_call: Callback to run when task is finished
        :return:
        """
        worker = TaskWorker(task_call)
        self._active_workers.add(worker)

        thread = QThread()
        worker.moveToThread(thread)

        # Connecter
        thread.started.connect(worker.run)
        worker.finished.connect(thread.quit)
        worker.finished.connect(self.finished)

        if finished_call:
            worker.finished.connect(finished_call)

        # Nettoyage automatique
        def cleanup():
            if worker in self._active_workers:
                self._active_workers.remove(worker)

            if thread.isRunning():
                thread.quit()
                thread.wait(50)

            QTimer.singleShot(0, thread.deleteLater)
            QTimer.singleShot(0, worker.deleteLater)

        worker.finished.connect(cleanup)

        thread.start()
        return worker


if __name__ == "__main__":

    def my_call():
        """
        Test call
        :return:
        """
        print("wait 2 seconds...")
        time.sleep(2)

    def on_finished_call():
        """
        Test call on finished task
        :return:
        """
        print("task finished.")

    def my_call2():
        """
        Test call
        :return:
        """
        print("call 2 wait 2 seconds...")
        time.sleep(2)

    def on_finished_call2():
        """
        Test call on finished task
        :return:
        """
        print("task 2 finished.")

    # test one worker
    app = QApplication(sys.argv)

    manager = ThreadManager(app)
    manager.start(my_call, on_finished_call)
    manager.start(my_call2, on_finished_call2)

    QTimer.singleShot(5000, app.quit)
    print("main thread activity...")

    app.exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import json

from tikka.adapters.network.datapod.connection import DataPodConnection
from tikka.adapters.network.datapod.profiles import DataPodProfiles
from tikka.domains.entities.datapod import DataPod
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.datapod.datapod import (
    NetworkDataPodException,
    NetworkDataPodInterface,
)


class NetworkDataPod(NetworkDataPodInterface):
    """
    NetworkDataPod class
    """

    def __init__(self):
        """
        Init NetworkDataPod instance
        """
        self._connection = DataPodConnection()
        self._profiles = DataPodProfiles(self)

    @property
    def connection(self) -> DataPodConnection:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkDataPodInterface.connection.__doc__
        )
        return self._connection

    def get(self) -> DataPod:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkDataPodInterface.get.__doc__
        )
        if not self.connection.is_connected() or self.connection.client is None:
            raise NetworkDataPodException(NetworkConnectionError())

        connection: DataPodConnection = self.connection
        if not connection.is_connected() or connection.client is None:
            raise NetworkDataPodException(NetworkConnectionError())
        try:
            result = connection.execute_query(endpoint="/g1/block/current/_source")
        except Exception as exception:
            raise NetworkDataPodException(exception)

        # {
        #          "version" : 10,
        #          "currency" : "g1",
        #          "number" : 29003,
        #          "issuer" : "H1yTj77m946f52u64FvR36o9SmD38Ye2j1H4XCvwBFXK",
        #          "hash" : "0000024FD141E2DA5E3BCAC15CD558EF47BD9CB38DE022F2E0E8727AA885FCD4",
        #          "medianTime" : 1498016342,
        #          "membersCount" : 161,
        #          "monetaryMass" : 10533000,
        #          "unitbase" : 0,
        #          "dividend" : null,
        #          "txCount" : 1,
        #          "txAmount" : 2500,
        #          "txChangeCount" : 0,
        #          "certCount" : 0
        #        }
        data = json.loads(result)

        return DataPod(self.connection.url, data["number"])

    def get_genesis_hash(self) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkDataPodInterface.get_genesis_hash.__doc__
        )
        if not self.connection.is_connected() or self.connection.client is None:
            raise NetworkDataPodException(NetworkConnectionError())

        connection: DataPodConnection = self.connection
        if not connection.is_connected() or connection.client is None:
            raise NetworkDataPodException(NetworkConnectionError())
        try:
            result = connection.execute_query(endpoint="/g1/block/0/_source")
        except Exception as exception:
            raise NetworkDataPodException(exception)

        #
        # {
        #          "version" : 10,
        #          "currency" : "g1",
        #          "number" : 29003,
        #          "issuer" : "H1yTj77m946f52u64FvR36o9SmD38Ye2j1H4XCvwBFXK",
        #          "hash" : "0000024FD141E2DA5E3BCAC15CD558EF47BD9CB38DE022F2E0E8727AA885FCD4",
        #          "medianTime" : 1498016342,
        #          "membersCount" : 161,
        #          "monetaryMass" : 10533000,
        #          "unitbase" : 0,
        #          "dividend" : null,
        #          "txCount" : 1,
        #          "txAmount" : 2500,
        #          "txChangeCount" : 0,
        #          "certCount" : 0
        #        }
        data = json.loads(result)

        return data["hash"].lower()

    @property
    def profiles(self) -> DataPodProfiles:
        """
        Return DataPodProfilesInterface instance

        :return:
        """
        return self._profiles



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import re
import sys
from typing import Optional

from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QFont, QMovie
from PyQt5.QtWidgets import QApplication, QDialog, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.identity import Identity, IdentityStatus
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
    ICON_LOADER,
)
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.worker import (
    ThreadManager,
)
from tikka.slots.pyqt.resources.gui.windows.identity_confirm_rc import (
    Ui_IdentityConfirmDialog,
)
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class IdentityConfirmWindow(QDialog, Ui_IdentityConfirmDialog):
    """
    IdentityConfirmWindow class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        parent: Optional[QWidget] = None,
    ):
        """
        Init identity confirm window

        :param application: Application instance
        :param account: Identity owner Account instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.event_bus = EventBus.instance()

        self.account = account
        self.identity: Optional[Identity] = self.application.identities.get_by_address(
            account.address
        )

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)
        self.accountValueLabel.setFont(self.monospace_font)

        # events
        self.nameLineEdit.textChanged.connect(self.on_name_line_edit_changed)
        self.confirmIdentityButton.clicked.connect(
            self._on_confirm_identity_button_clicked
        )

        ##############################
        # ASYNC METHODS
        ##############################
        self.thread_manager: ThreadManager = ThreadManager(self)

        # debounce timers
        self.name_debounce_timer = QTimer()
        self.name_debounce_timer.timeout.connect(self._name_debounce_call)

        self.confirmIdentityButton.setEnabled(False)
        self.update_ui()
        self.nameLineEdit.setFocus()

    def on_name_line_edit_changed(self) -> None:
        """
        Triggered when name line edit is changed

        :return:
        """
        if self.name_debounce_timer.isActive():
            self.name_debounce_timer.stop()
        self.name_debounce_timer.start(DEBOUNCE_TIME)

    def _name_debounce_call(self):
        """
        Debounce function triggered only after idle time when typing name

        :return:
        """
        # Stop name_debounce_timer to avoid infinite loop
        if self.name_debounce_timer.isActive():
            self.name_debounce_timer.stop()

        name = self.nameLineEdit.text().strip()

        # Check name validity
        # Criteria:
        # 1. Length <= 42 characters
        # 2. Only ASCII alphanumeric characters or '_' or '-'
        # 3. Cannot start or end with '-' or '_' (optional, comment if not needed)

        # Define the regex pattern
        # ^[a-zA-Z0-9] - must start with alphanumeric
        # [a-zA-Z0-9_-]* - zero or more alphanumeric, underscore, or hyphen
        # $ - end of string
        # Total length <= 42
        pattern = r"^[a-zA-Z0-9][a-zA-Z0-9_-]{0,41}$"

        # Check if name is valid
        name_validity = bool(re.match(pattern, name)) and len(name) <= 42

        if not name_validity:
            self.errorLabel.setText(self._("Invalid name"))
            self.confirmIdentityButton.setEnabled(False)
        else:
            self.errorLabel.setText("")
            if self.identity.status == IdentityStatus.UNCONFIRMED:
                self.confirmIdentityButton.setEnabled(True)

    def update_ui(self):
        """
        Update UI fields

        :return:
        """
        self.accountValueLabel.setText(self.account.address)
        if self.identity is not None:
            display_identity_status = {
                IdentityStatus.UNCONFIRMED.value: self._("Unconfirmed"),
                IdentityStatus.UNVALIDATED.value: self._("Unvalidated"),
                IdentityStatus.MEMBER.value: self._("Member"),
                IdentityStatus.NOT_MEMBER.value: self._("Not member"),
                IdentityStatus.REVOKED.value: self._("Revoked"),
            }
            identity_status_color = {
                IdentityStatus.UNCONFIRMED.value: "orange",
                IdentityStatus.UNVALIDATED.value: "orange",
                IdentityStatus.MEMBER.value: "green",
                IdentityStatus.NOT_MEMBER.value: "red",
                IdentityStatus.REVOKED.value: "red",
            }
            identity_status_style = f"""
                                    color: white;
                                    border: 3px solid "{identity_status_color[self.identity.status.value]}";
                                    border-radius: 10%;
                                    background: "{identity_status_color[self.identity.status.value]}";
                                    """
            self.identityIndexValueLabel.setText(f"#{self.identity.index}")
            self.identityStatusValueLabel.setText(
                display_identity_status[self.identity.status.value]
            )
            self.identityStatusValueLabel.setStyleSheet(identity_status_style)

    def _on_confirm_identity_button_clicked(self):
        """
        Run when use click confirm identity button

        :return:
        """
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.loaderIconLabel.show()
        self.confirmIdentityButton.setDisabled(True)
        self.thread_manager.start(
            self.network_confirm_identity, self._on_finished_network_confirm_identity
        )

    def network_confirm_identity(self):
        """
        Confirm identity on the network

        :return:
        """
        name = self.nameLineEdit.text().strip()
        try:
            self.application.identities.network_confirm(
                self.application.wallets.get_keypair(self.account.address),
                name,
            )
        except Exception as exception:
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(str(exception))
            logging.exception(exception)
        else:
            self.errorLabel.setStyleSheet("color: green;")
            self.errorLabel.setText(
                self._("Identity {name}#{index} confirmed!").format(
                    name=name, index=self.identity.index
                )
            )
            try:
                self.application.identities.network_update_identity(
                    self.account.address
                )
            except Exception as exception:
                logging.exception(exception)

    def _on_finished_network_confirm_identity(self):
        """
        When async network_create_identity is finished

        :return:
        """
        self.identity = self.application.identities.get_by_address(self.account.address)
        self.confirmIdentityButton.setDisabled(False)
        self.loaderIconLabel.hide()
        self.update_ui()
        if self.identity.status == IdentityStatus.UNVALIDATED:
            self.confirmIdentityButton.setEnabled(False)
            self.event_bus.identity_confirmed.emit(self.identity)
            logging.debug(
                "identity_confirmed signal emitted from identity_confirm window %s#%s",
                self.identity.name,
                self.identity.index,
            )
        else:
            self.confirmIdentityButton.setDisabled(False)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    account_ = Account("xxxx")
    IdentityConfirmWindow(application_, account_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Dict, List, Optional, OrderedDict

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont, QMovie
from PyQt5.QtWidgets import QApplication, QCompleter, QDialog, QWidget
from scalecodec.utils.ss58 import is_valid_ss58_address

from tikka.domains.application import Application
from tikka.domains.entities.account import Account, AccountCryptoType
from tikka.domains.entities.address import DisplayAddress
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.libs.keypair import Keypair
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_LOADER,
    SELECTED_UNIT_PREFERENCES_KEY,
    TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY,
    TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.worker import ThreadManager
from tikka.slots.pyqt.resources.gui.windows.transfer_rc import Ui_TransferDialog
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class TransferWindow(QDialog, Ui_TransferDialog):
    """
    TransferWindow class
    """

    def __init__(
        self,
        application: Application,
        sender: Optional[Account] = None,
        recipient: Optional[Account] = None,
        parent: Optional[QWidget] = None,
    ):
        """
        Init transfer window

        :param application: Application instance
        :param sender: Sender Account instance or None
        :param recipient: Recipient Account instance or None
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.event_bus = EventBus.instance()
        self.event_bus.account_balances_updated.connect(
            self.on_account_balances_updated
        )

        preferences_sender_address = self.application.repository.preferences.get(
            TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY
        )
        if sender is None and preferences_sender_address is not None:
            preferences_sender_account = self.application.accounts.get_by_address(
                preferences_sender_address
            )
            if preferences_sender_account is not None:
                sender = preferences_sender_account

        preferences_recipient_address = self.application.repository.preferences.get(
            TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY
        )
        if recipient is None and preferences_recipient_address is not None:
            preferences_recipient_account = self.application.accounts.get_by_address(
                preferences_recipient_address
            )
            if preferences_recipient_account is not None:
                recipient = preferences_recipient_account

        self.sender_account = sender
        self.recipient_account = recipient
        self._ = self.application.translator.gettext

        # substrate error message for gettext extraction
        self._("Transfer/payment would kill account")
        self._("Balance too low to send value")

        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)

        # init unit from preference or default
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.unit = unit_preference
        else:
            self.unit = AMOUNT_UNIT_KEY

        self.init_units()

        self.amount_value = 0
        self.fees = None
        self.transfer_success: Optional[bool] = None

        if self.sender_account is not None:
            if self.sender_account.name is None:
                self.senderAddressValueLabel.setText("")
            else:
                self.senderAddressValueLabel.setText(self.sender_account.address)
                self.senderAddressValueLabel.setFont(self.monospace_font)

        if self.recipient_account is not None:
            if self.recipient_account.name is None:
                self.recipientNameOrAddressLineEdit.setText(
                    self.recipient_account.address
                )
                self.recipientNameOrAddressLineEdit.setFont(self.monospace_font)
                self.recipientNameOrAddressValueLabel.setText("")
            else:
                self.recipientNameOrAddressLineEdit.setText(self.recipient_account.name)
                self.recipientNameOrAddressLineEdit.setFont(QFont())
                self.recipientNameOrAddressValueLabel.setText(
                    self.recipient_account.address
                )
                self.recipientNameOrAddressValueLabel.setFont(self.monospace_font)

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        # autocomplete  (dirty version)
        # todo: create a model to fetch only a filtered list from database on edit
        self.account_address_by_name: Dict[str, str] = {}
        account_address_by_name_duplicate_count = 2
        for account_instance in self.application.accounts.get_list():
            if (
                account_instance.name in self.account_address_by_name
                and account_instance.name is not None
            ):
                self.account_address_by_name[
                    f"{account_instance.name} ({account_address_by_name_duplicate_count})"
                ] = account_instance.address
                account_address_by_name_duplicate_count += 1
            else:
                if account_instance.name is None:
                    self.account_address_by_name[
                        account_instance.address
                    ] = account_instance.address
                else:
                    self.account_address_by_name[
                        account_instance.name
                    ] = account_instance.address

        identities_by_address = self.application.identities.get_by_addresses(
            list(self.account_address_by_name.values())
        )
        for address, identity in identities_by_address.items():
            if identity is not None:
                if identity.name in self.account_address_by_name:
                    self.account_address_by_name[
                        f"{identity.name} ({account_address_by_name_duplicate_count})"
                    ] = address
                    account_address_by_name_duplicate_count += 1
                if identity.name is not None:
                    self.account_address_by_name[identity.name] = address

        # recipient autocomplete
        self.recipient_account_wordlist = list(
            self.account_address_by_name.keys()
        ) + list(self.account_address_by_name.values())
        self.recipient_account_wordlist = list(
            OrderedDict.fromkeys(self.recipient_account_wordlist)
        )
        recipient_completer = QCompleter(self.recipient_account_wordlist)
        self.recipientNameOrAddressLineEdit.setCompleter(recipient_completer)
        self.recipientNameOrAddressLineEdit.completer().setCaseSensitivity(
            Qt.CaseInsensitive
        )

        self.init_sender_combo_box()

        self._update_ui()

        # events
        self.senderComboBox.activated.connect(self._on_sender_combo_box_changed)
        self.recipientNameOrAddressLineEdit.textChanged.connect(
            self._on_recipient_address_line_edit_changed
        )
        self.amountDoubleSpinBox.valueChanged.connect(
            self._on_amount_double_spin_box_changed
        )
        self.amountUnitComboBox.activated.connect(self._on_unit_changed)
        self.feesButton.clicked.connect(self._on_fees_button_clicked)

        self.allCheckBox.toggled.connect(self._on_all_check_box_toggled)
        self.sendButton.clicked.connect(self._on_send_button_clicked)
        self.buttonBox.button(self.buttonBox.Close).clicked.connect(self.close)

        ##############################
        # ASYNC METHODS
        ##############################
        self.thread_manager: ThreadManager = ThreadManager(self)

        # if receiver account is unknown...
        if (
            self.recipient_account is not None
            and self.application.accounts.get_by_address(self.recipient_account.address)
            is None
        ):
            # get balance from node
            self.thread_manager.start(
                self.fetch_recipient_balance_from_network,
                self._on_finished_fetch_recipient_balance_from_network,
            )

    def init_sender_combo_box(self) -> None:
        """
        Init combobox with accounts having a wallet
        """
        self.senderComboBox.clear()
        self.senderComboBox.addItem("?", userData=None)

        identities = {
            identity.address: identity
            for identity in self.application.identities.list()
        }

        for account in self.application.accounts.get_list():
            if self.application.wallets.exists(account.address):
                account_name = f" - {account.name}" if account.name is not None else ""  # type: ignore
                identity_name = ""
                identity = identities.get(account.address)
                if identity is not None:
                    identity_name = f" - {identity.name}#{identity.index}"
                self.senderComboBox.addItem(
                    f"{DisplayAddress(account.address).shorten}{identity_name}{account_name}",
                    userData=account.address,
                )

        if self.sender_account is not None:
            sender_index = self.senderComboBox.findData(self.sender_account.address)
            if sender_index > -1:
                self.senderComboBox.setCurrentIndex(sender_index)

    def _on_unit_changed(self):
        """
        Triggered when unit combo box is changed

        :return:
        """
        self.unit = self.amountUnitComboBox.currentData()
        self._update_ui()

    def _on_amount_double_spin_box_changed(self):
        """
        Triggered when the amount spin box is changed

        :return:
        """
        self.amount_value = self.amountDoubleSpinBox.value()
        self._update_ui()

    def _on_all_check_box_toggled(self, checked: bool):
        """
        Triggered when the amount spin box is changed

        :return:
        """
        self.amountDoubleSpinBox.setDisabled(checked)
        self._update_ui()

    def fetch_recipient_balance_from_network(self):
        """
        Fetch last account data from the network

        :return:
        """
        self.loaderIconLabel.show()
        try:
            # if account not in stored accounts...
            if (
                self.application.accounts.get_by_address(self.recipient_account.address)
                is None
            ):
                recipient_balance = self.application.accounts.network_get_balance(
                    self.recipient_account.address
                )
                if recipient_balance is not None:
                    self.recipient_account.balance = recipient_balance.total
                    self.recipient_account.balance_available = (
                        recipient_balance.available
                    )
                    self.recipient_account.balance_reserved = recipient_balance.reserved
            else:
                self.recipient_account = (
                    self.application.accounts.network_update_balance(
                        self.recipient_account
                    )
                )
        except Exception as exception:
            # keep balance from cache DB if network failure
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_finished_fetch_recipient_balance_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self._update_ui()

    def fetch_sender_balance_from_network(self):
        """
        Fetch last sender account data from the network

        :return:
        """
        self.loaderIconLabel.show()

        try:
            self.sender_account = self.application.accounts.network_update_balance(
                self.sender_account
            )
        except Exception as exception:
            # keep balance from cache DB if network failure
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_finished_fetch_sender_balance_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self._update_ui()

    def _on_sender_combo_box_changed(self) -> None:
        """
        Triggered when the sender combobox is changed

        :return:
        """
        self.errorLabel.setText("")
        address = self.senderComboBox.currentData()
        if address is None:
            self.senderAddressValueLabel.setText("")
            self.senderBalanceValueLabel.setText("?")
            self.sender_account = None
            return None
        try:
            address_is_valid = is_valid_ss58_address(address)
        except IndexError:
            address_is_valid = False

        if not address_is_valid:
            self.senderAddressValueLabel.setText("")
            self.senderBalanceValueLabel.setText("?")
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Invalid address! Please check it again."))
            self.sender_account = None
        else:
            self._get_sender_account(address)

        self._update_ui()

        return None

    def _on_recipient_address_line_edit_changed(self):
        """
        Triggered when text in the recipient address field is changed

        :return:
        """
        self.errorLabel.setText("")
        address_or_name = self.recipientNameOrAddressLineEdit.text().strip()

        try:
            address_is_valid = is_valid_ss58_address(address_or_name)
        except IndexError:
            address_is_valid = False

        if not address_is_valid:
            # if entry is a known account name...
            if address_or_name in self.account_address_by_name:
                self.recipientNameOrAddressLineEdit.setFont(QFont())
                # display address under the name
                self.recipientNameOrAddressValueLabel.setFont(self.monospace_font)
                self.recipientNameOrAddressValueLabel.setText(
                    self.account_address_by_name[address_or_name]
                )
                self._get_recipient_account(
                    self.account_address_by_name[address_or_name]
                )
            else:
                self.recipientNameOrAddressValueLabel.setText("")
                self.recipientBalanceValueLabel.setText("?")
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(
                    self._("Invalid address! Please check it again.")
                )
                self.recipient_account = None
        else:
            self.recipientNameOrAddressLineEdit.setFont(self.monospace_font)
            # display name under the address
            for key, value in self.account_address_by_name.items():
                if value == address_or_name:
                    self.recipientNameOrAddressValueLabel.setFont(QFont())
                    self.recipientNameOrAddressValueLabel.setText(key)
            self._get_recipient_account(address_or_name)

        self._update_ui()

    def _get_sender_account(self, address: str) -> None:
        """
        Get sender account

        :return:
        """
        self.sender_account = None
        try:
            Keypair(
                ss58_address=address,
                ss58_format=self.application.currencies.get_current().ss58_format,
            )
        except ValueError as exception:
            logging.exception(exception)
            self.transfer_success = False
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Invalid address! Please check it again."))
            self.sender_account = None
            return None

        # save in preferences
        self.application.repository.preferences.set(
            TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY, address
        )

        # search in local account list
        for account in self.application.accounts.get_list():
            if account.address == address:
                self.sender_account = account
                self._update_ui()
                return None

        # create account
        self.sender_account = Account(address)
        # fetch balance
        self.thread_manager.start(
            self.fetch_sender_balance_from_network,
            self._on_finished_fetch_sender_balance_from_network,
        )

        return None

    def _get_recipient_account(self, address: str) -> None:
        """
        Get recipient account

        :return:
        """
        self.recipient_account = None
        try:
            Keypair(
                ss58_address=address,
                ss58_format=self.application.currencies.get_current().ss58_format,
            )
        except ValueError as exception:
            logging.exception(exception)
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Invalid address! Please check it again."))
            self.recipient_account = None
            return None

        # save in preferences
        self.application.repository.preferences.set(
            TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY, address
        )

        # search in local account list
        for account in self.application.accounts.get_list():
            if account.address == address:
                self.recipient_account = account
                self._update_ui()
                return None

        # create account
        self.recipient_account = Account(address)
        # fetch balance
        self.thread_manager.start(
            self.fetch_recipient_balance_from_network,
            self._on_finished_fetch_recipient_balance_from_network,
        )

        return None

    def _update_ui(self):
        """
        Update UI

        :return:
        """
        if self.sender_account is not None:
            stored_account = self.application.accounts.get_by_address(
                self.sender_account.address
            )
            # update only stored account
            if stored_account is not None:
                self.sender_account = stored_account

        if self.recipient_account is not None:
            stored_account = self.application.accounts.get_by_address(
                self.recipient_account.address
            )
            # update only stored account
            if stored_account is not None:
                self.recipient_account = stored_account

        # balances amount
        amount = self.application.amounts.get_amount(self.unit)
        if self.sender_account is not None:
            self.senderAddressValueLabel.setText(self.sender_account.address)
            if self.sender_account.balance is None:
                self.senderBalanceValueLabel.setText("?")
            else:
                self.senderBalanceValueLabel.setText(
                    self.locale().toCurrencyString(
                        amount.value(self.sender_account.balance), amount.symbol()
                    )
                )
                display_sender_reserved_balance = self.locale().toCurrencyString(
                    amount.value(self.sender_account.balance_reserved), amount.symbol()
                )
                self.senderBalanceReservedValueLabel.setText(
                    f"[-{display_sender_reserved_balance}]"
                )

                self.amountDoubleSpinBox.setMaximum(
                    amount.value(
                        self.sender_account.balance
                        - self.sender_account.balance_reserved
                    )
                )

        if self.recipient_account is not None:
            if self.recipient_account.balance is None:
                self.recipientBalanceValueLabel.setText("?")
                self.recipientBalanceValueLabel.setToolTip(
                    self._(
                        "Account balance unknown! Send only one unit and make sure the owner can get it"
                    )
                )
            else:
                self.recipientBalanceValueLabel.setText(
                    self.locale().toCurrencyString(
                        amount.value(self.recipient_account.balance), amount.symbol()
                    )
                )
                self.recipientBalanceValueLabel.setToolTip("")

        # fees amount
        if self.fees is not None:
            self.feesValueLabel.setText(
                self.locale().toCurrencyString(amount.value(self.fees), amount.symbol())
            )
        else:
            self.feesValueLabel.setText("?")

        # buttons
        if (
            self.sender_account is not None
            and self.sender_account.balance is not None
            and self.recipient_account is not None
            and (self.amount_value > 0 or self.allCheckBox.isChecked())
        ):
            self.sendButton.setDisabled(False)
            self.feesButton.setDisabled(False)
        else:
            self.feesButton.setDisabled(True)
            self.sendButton.setDisabled(True)

    def init_units(self) -> None:
        """
        Init units combobox for transfer amount

        :return:
        """
        self.amountUnitComboBox.clear()

        for key, amount in self.application.amounts.register.items():
            self.amountUnitComboBox.addItem(amount.name(), userData=key)
        self.amountUnitComboBox.setCurrentIndex(
            self.amountUnitComboBox.findData(self.unit)
        )

    def _on_fees_button_clicked(self):
        """
        Triggered when user click on Fees button

        :return:
        """
        # sender account locked...
        if not self.application.wallets.is_unlocked(self.sender_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.sender_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.feesButton.setDisabled(True)
        self.loaderIconLabel.show()

        self.thread_manager.start(
            self.fetch_fees_from_network, self._on_finished_fetch_fees_from_network
        )

    def fetch_fees_from_network(self):
        """
        Fetch fees amount from network

        :return:
        """
        amount = self.application.amounts.get_amount(self.unit)
        # get value as blockchain units
        blockchain_value = amount.blockchain_value(self.amountDoubleSpinBox.value())
        sender_keypair = self.application.wallets.get_keypair(
            self.sender_account.address
        )
        try:
            self.fees = self.application.transfers.network_fees(
                sender_keypair, self.recipient_account.address, blockchain_value
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_finished_fetch_fees_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.feesButton.setDisabled(False)
        self._update_ui()

    def _on_send_button_clicked(self):
        """
        Triggered when user click on Send button

        :return:
        """
        # sender account locked...
        if not self.application.wallets.is_unlocked(self.sender_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.sender_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.sendButton.setDisabled(True)
        self.loaderIconLabel.show()
        # send transfer to network
        self.thread_manager.start(
            self.send_transfer_to_network, self._on_finished_send_transfer_to_network
        )

    def send_transfer_to_network(self):
        """
        Send transfer to network

        :return:
        """
        self.errorLabel.setText("")
        # get value as blockchain units
        amount = self.application.amounts.get_amount(self.unit)
        blockchain_value = amount.blockchain_value(self.amountDoubleSpinBox.value())
        sender_keypair = self.application.wallets.get_keypair(
            self.sender_account.address
        )
        comment: Optional[str] = (
            self.commentLineEdit.text().strip()
            if self.commentLineEdit.text().strip() != ""
            else None
        )

        if sender_keypair is None:
            return

        # if amount needs unclaimed UDs funds...
        if (
            blockchain_value > self.sender_account.balance_available
            or self.allCheckBox.isChecked()
        ):
            identity = self.application.identities.get_by_address(
                self.sender_account.address
            )
            if identity is not None:
                self.errorLabel.setStyleSheet("color: green;")
                self.errorLabel.setText(self._("Claiming Universal Dividends..."))
                # get unclaimed UDs
                try:
                    self.application.identities.network_claim_uds(sender_keypair)
                except Exception as exception:
                    self.transfer_success = False
                    self.errorLabel.setStyleSheet("color: red;")
                    self.errorLabel.setText(self._(str(exception)))
                    logging.exception(exception)

        try:
            if self.allCheckBox.isChecked():
                extrinsic_receipt = self.application.transfers.network_send_all(
                    sender_keypair, self.recipient_account.address, False, comment
                )
            else:
                extrinsic_receipt = self.application.transfers.network_send(
                    sender_keypair,
                    self.recipient_account.address,
                    blockchain_value,
                    comment,
                )
        except Exception as exception:
            self.transfer_success = False
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)
        else:
            if extrinsic_receipt is None:
                self.transfer_success = False
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(
                    self._("Transfer failed. Please check logs to understand why")
                )
            elif extrinsic_receipt.is_success is False:
                self.transfer_success = False
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(self._(extrinsic_receipt.error_message["name"]))
            else:
                self.transfer_success = True
                self.errorLabel.setStyleSheet("color: green;")
                self.errorLabel.setText(self._("Transfer done"))

    def _on_finished_send_transfer_to_network(self):
        """
        Triggered when async request send_transfer_to_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.sendButton.setDisabled(False)
        self.event_bus.transfer_sent.emit([self.sender_account, self.recipient_account])

    def on_account_balances_updated(self, accounts: List[Account]):
        """
        Triggered when account balances has been updated

        :return:
        """
        if self.sender_account is not None:
            if self.sender_account.address in [account.address for account in accounts]:
                self._update_ui()
                return
        if self.recipient_account is not None:
            if self.recipient_account.address in [
                account.address for account in accounts
            ]:
                self._update_ui()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)

    account1 = application_.accounts.create_new_root_account(
        "album cute glance oppose hub fury strategy health dust rebuild trophy magic",
        "en",
        AccountCryptoType.ED25519,
        "test account",
        "aaaaaa",
    )
    if account1 is not None:
        account1.balance = 1000
    account2 = application_.accounts.create_new_account(
        "album cute glance oppose hub fury strategy health dust rebuild trophy magic",
        "en",
        "//1",
        AccountCryptoType.ED25519,
        "test account",
        "aaaaaa",
    )
    if account2 is not None:
        account2.balance = 0

    TransferWindow(application_, account1, account2).exec_()
